import logging
import os
from itertools import zip_longest

from pydm.utilities import IconFont, remove_protocol
from pydm.widgets.base import PyDMPrimitiveWidget
from pydm.widgets.channel import PyDMChannel
from pydm.widgets.embedded_display import PyDMEmbeddedDisplay
from qtpy.QtCore import Q_ENUMS, Property, QSize
from qtpy.QtGui import QCursor, QPainter
from qtpy.QtWidgets import (QFrame, QHBoxLayout, QSizePolicy, QStyle,
                            QStyleOption, QTabWidget, QVBoxLayout, QWidget)

from ..utils import refresh_style

logger = logging.getLogger(__name__)


class ContentLocation:
    """
    Enum Class to be used by the widgets to configure the Controls Content
    Location.
    """
    Hidden = 0
    Top = 1
    Bottom = 2
    Left = 3
    Right = 4


class PCDSSymbolBase(QWidget, PyDMPrimitiveWidget, ContentLocation):
    """
    Base class to be used for all PCDS Symbols.

    Parameters
    ----------
    parent : QWidget
        The parent widget for this symbol.
    """

    _qt_designer_ = {
        "group": "PCDS Symbols",
        "is_container": False,
    }

    EXPERT_OPHYD_CLASS = ""

    Q_ENUMS(ContentLocation)
    ContentLocation = ContentLocation

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self._expert_display = None
        self.interlock = None
        self._channels_prefix = None
        self._rotate_icon = False

        self._show_icon = True
        self._show_status_tooltip = True
        self._icon_size = -1
        self._icon = None

        self._icon_cursor = self.setCursor(
            QCursor(IconFont().icon("file").pixmap(16, 16))
        )

        self._expert_ophyd_class = self.EXPERT_OPHYD_CLASS or ""

        self.interlock = QFrame(self)
        self.interlock.setObjectName("interlock")
        self.interlock.setSizePolicy(QSizePolicy.Expanding,
                                     QSizePolicy.Expanding)

        self.controls_frame = QFrame(self)
        self.controls_frame.setObjectName("controls")
        self.controls_frame.setSizePolicy(QSizePolicy.Maximum,
                                          QSizePolicy.Maximum)
        self.setLayout(QVBoxLayout())
        self.layout().setSpacing(0)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self.interlock)
        if not hasattr(self, '_controls_location'):
            self._controls_location = ContentLocation.Bottom
        self.setup_icon()
        self.assemble_layout()
        self.update_status_tooltip()
        self.ui_file_paths = []

        self.ui_file_macros = []
        self.ui_file_titles = []

        self.tab_widget = None
        self.embedded_displays = list()

    def sizeHint(self):
        """
        Suggested initial size for the widget.

        Returns
        -------
        size : QSize
        """
        return QSize(200, 200)

    @Property(ContentLocation)
    def controlsLocation(self):
        """
        Property controlling where the controls frame will be displayed.

        Returns
        -------
        location : ContentLocation
        """
        return self._controls_location

    @controlsLocation.setter
    def controlsLocation(self, location):
        """
        Property controlling where the controls frame will be displayed.

        Parameters
        ----------
        location : ContentLocation
        """
        if location != self._controls_location:
            self._controls_location = location
            self.assemble_layout()

    @Property(str)
    def channelsPrefix(self):
        """
        The prefix to be used when composing the channels for each of the
        elements of the symbol widget.

        The prefix must include the protocol as well. E.g.: ca://VALVE

        Returns
        -------
        str
        """
        return self._channels_prefix

    @channelsPrefix.setter
    def channelsPrefix(self, prefix):
        """
        The prefix to be used when composing the channels for each of the
        elements of the symbol widget.

        The prefix must include the protocol as well. E.g.: ca://VALVE

        Parameters
        ----------
        prefix : str
            The prefix to be used for the channels.
        """
        if prefix != self._channels_prefix:
            self._channels_prefix = prefix
            self.destroy_channels()
            self.create_channels()

    @property
    def icon(self):
        return self._icon

    @icon.setter
    def icon(self, icon):
        if self._icon != icon:
            self._icon = icon
            self.setup_icon()
            self.iconSize = self.iconSize
            self.assemble_layout()

    @Property(bool)
    def showIcon(self):
        """
        Whether or not to show the symbol icon when rendering the widget.

        Returns
        -------
        bool
        """
        return self._show_icon

    @showIcon.setter
    def showIcon(self, value):
        """
        Whether or not to show the symbol icon when rendering the widget.

        Parameters
        ----------
        value : bool
            Shows the Icon if True, hides it otherwise.
        """
        if value != self._show_icon:
            self._show_icon = value
            if self.icon:
                self.icon.setVisible(self._show_icon)
            self.assemble_layout()

    @Property(bool)
    def showStatusTooltip(self):
        """
        Whether or not to show a detailed status tooltip including the state
        of the widget components such as Interlock, Error, State and more.

        Returns
        -------
        bool
        """
        return self._show_status_tooltip

    @showStatusTooltip.setter
    def showStatusTooltip(self, value):
        """
        Whether or not to show a detailed status tooltip including the state
        of the widget components such as Interlock, Error, State and more.

        Parameters
        ----------
        value : bool
            Displays the tooltip if True.

        """
        if value != self._show_status_tooltip:
            self._show_status_tooltip = value

    @Property(int)
    def iconSize(self):
        """
        The size of the icon in pixels.

        Returns
        -------
        int
        """
        return self._icon_size

    @iconSize.setter
    def iconSize(self, size):
        """
        The size of the icon in pixels.

        Parameters
        ----------
        size : int
            A value > 0 will constrain the size of the icon to the defined
            value.
            If the value is <= 0 it will expand to fill the space available.

        """
        if not self.icon:
            return

        if size <= 0:
            size = - 1
            min_size = 1
            max_size = 999999
            self.icon.setSizePolicy(QSizePolicy.Expanding,
                                    QSizePolicy.Expanding)
            self.icon.setMinimumSize(min_size, min_size)
            self.icon.setMaximumSize(max_size, max_size)

        else:
            self.icon.setFixedSize(size, size)
            self.icon.setSizePolicy(QSizePolicy.Fixed,
                                    QSizePolicy.Fixed)

        self._icon_size = size
        self.icon.update()

    @Property(bool)
    def rotateIcon(self):
        """
        Rotate the icon 90 degrees clockwise

        Returns
        -------
        rotate : bool
        """
        return self._rotate_icon

    @rotateIcon.setter
    def rotateIcon(self, rotate):
        """
        Rotate the icon 90 degrees clockwise

        Parameters
        ----------
        rotate : bool
        """
        self._rotate_icon = rotate
        angle = 90 if self._rotate_icon else 0
        if self.icon:
            self.icon.rotation = angle

    @Property(str)
    def expertOphydClass(self):
        """
        The full qualified name of the Ophyd class to be used for the Expert
        screen to be generated using Typhos.

        Returns
        -------
        str
        """
        klass = self._expert_ophyd_class
        if isinstance(klass, type):
            return f"{klass.__module__}.{klass.__name__}"
        return klass

    @expertOphydClass.setter
    def expertOphydClass(self, klass):
        """
        The full qualified name of the Ophyd class to be used for the Expert
        screen to be generated using Typhos.

        Parameters
        ----------
        klass : bool
        """
        if self.expertOphydClass != klass:
            self._expert_ophyd_class = klass

    def paintEvent(self, evt):
        """
        Paint events are sent to widgets that need to update themselves,
        for instance when part of a widget is exposed because a covering
        widget was moved.

        This method handles the painting with parameters from the stylesheet.

        Parameters
        ----------
        evt : QPaintEvent
        """
        painter = QPainter(self)
        opt = QStyleOption()
        opt.initFrom(self)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)
        painter.setRenderHint(QPainter.Antialiasing)
        super().paintEvent(evt)

    def clear(self):
        """
        Remove all inner widgets from the interlock frame layout.
        """
        if not self.interlock:
            return
        layout = self.interlock.layout()
        if layout is None:
            return
        while layout.count() != 0:
            item = layout.itemAt(0)
            if item is not None:
                layout.removeItem(item)

        # Trick to remove the existing layout by re-parenting it in an
        # empty widget.
        QWidget().setLayout(self.interlock.layout())

    def assemble_layout(self):
        """
        Assembles the widget's inner layout depending on the ContentLocation
        and other configurations set.

        """
        if not self.interlock:
            return
        self.clear()

        # (Layout, items)
        widget_map = {
            ContentLocation.Hidden: (QVBoxLayout,
                                     [self.icon]),
            ContentLocation.Top: (QVBoxLayout,
                                  [self.controls_frame,
                                   self.icon]),
            ContentLocation.Bottom: (QVBoxLayout,
                                     [self.icon,
                                      self.controls_frame]),
            ContentLocation.Left: (QHBoxLayout,
                                   [self.controls_frame,
                                    self.icon]),
            ContentLocation.Right: (QHBoxLayout,
                                    [self.icon,
                                     self.controls_frame]),
        }

        layout = widget_map[self._controls_location][0]()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        self.interlock.setLayout(layout)

        widgets = widget_map[self._controls_location][1]

        # Hide the controls box if they are not going to be included in layout
        controls_visible = self._controls_location != ContentLocation.Hidden
        self.controls_frame.setVisible(controls_visible)

        for widget in widgets:
            if widget is None:
                continue
            # Each widget is in a separate layout to help with expansion rules
            box_layout = QHBoxLayout()
            box_layout.addWidget(widget)
            layout.addLayout(box_layout)

    def setup_icon(self):
        if not self.icon:
            return
        self.icon.setMinimumSize(16, 16)
        self.icon.setSizePolicy(QSizePolicy.Expanding,
                                QSizePolicy.Expanding)
        self.icon.setVisible(self._show_icon)
        self.iconSize = 32
        if hasattr(self.icon, 'clicked'):
            self.icon.clicked.connect(self._handle_icon_click)
            if self._expert_display is not None:
                self.icon.setCursor(self._icon_cursor)

    def _handle_icon_click(self):
        if not self.channelsPrefix:
            logger.error('No channel prefix specified.'
                         'Cannot proceed with opening expert screen for %s.',
                         self.__class__.__name__)
            return

        if self.tab_widget is not None:
            logger.debug('Bringing existing custom display to front.')
            self.tab_widget.show()
            self.tab_widget.raise_()
            return
        elif self._expert_display is not None:
            logger.debug('Bringing existing display to front.')
            self._expert_display.show()
            self._expert_display.raise_()
            return

        prefix = remove_protocol(self.channelsPrefix)
        klass = self.expertOphydClass
        if not klass:
            logger.error('No expertOphydClass specified for pcdswidgets %s',
                         self.__class__.__name__)
            return
        name = prefix.replace(':', '_')

        try:
            import typhos
        except ImportError:
            logger.error('Typhos not installed. Cannot create display.')
            return

        kwargs = {"name": name, "prefix": prefix}
        display = typhos.TyphosDeviceDisplay.from_class(klass, **kwargs)
        self._expert_display = display
        display.destroyed.connect(self._cleanup_expert_display)

        if len(self.ui_file_paths) > 0:
            self.tab_widget = QTabWidget()
            self.tab_widget.setTabPosition(QTabWidget.TabPosition.West)
            self.tab_widget.addTab(display, "Typhos")

            for file_path, title, macros in zip_longest(
                self.ui_file_paths,
                self.ui_file_titles,
                self.ui_file_macros
            ):
                embedded = PyDMEmbeddedDisplay()
                title = title or file_path
                macros = macros or ''

                embedded.set_macros_and_filename(file_path, macros)
                self.tab_widget.addTab(embedded, title)
                self.embedded_displays.append(embedded)

            self.tab_widget.show()

        elif display:
            display.show()

    @Property('QStringList')
    def ui_paths(self):
        return self.ui_file_paths

    @ui_paths.setter
    def ui_paths(self, path):
        if path != self.ui_file_paths:
            self.ui_file_paths = path

    @Property('QStringList')
    def ui_macros(self):
        return self.ui_file_macros

    @ui_macros.setter
    def ui_macros(self, macros):
        if macros != self.ui_macros:
            self.ui_file_macros = macros

    @Property('QStringList')
    def ui_titles(self):
        return self.ui_file_titles

    @ui_titles.setter
    def ui_titles(self, titles):
        if titles != self.ui_file_titles:
            self.ui_file_titles = titles

    def _cleanup_expert_display(self, *args, **kwargs):
        self._expert_display = None

    def status_tooltip(self):
        """
        Assemble and returns the status tooltip for the symbol.

        Returns
        -------
        str
        """
        status = ""
        if hasattr(self, 'NAME'):
            status = self.NAME
        if status:
            status += os.linesep
        status += f"PV Prefix: {self.channelsPrefix}"
        return status

    def destroy_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        destroyed.
        """
        for v in self.__dict__.values():
            if isinstance(v, PyDMChannel):
                v.disconnect()

    def create_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        created.
        This method must be implemented on the subclasses and mixins as needed.
        By default this method does nothing.
        """
        pass

    def update_stylesheet(self):
        """
        Invoke the stylesheet update process on the widget and child widgets to
        reflect changes on the properties.
        """
        refresh_style(self)

    def update_status_tooltip(self):
        """
        Set the tooltip on the symbol to the content of status_tooltip.
        """
        self.setToolTip(self.status_tooltip())



__all__ = ['HotCathodeGauge', 'RoughGauge', 'ColdCathodeGauge', 'IonPump',
           'TurboPump', 'ScrollPump', 'GetterPump', 'RGA', 'PneumaticValve',
           'ApertureValve', 'FastShutter', 'NeedleValve', 'ProportionalValve',
           'RightAngleManualValve', 'ControlValve', 'ControlOnlyValveNC',
           'ControlOnlyValveNO', 'PneumaticValveNO', 'PneumaticValveDA']

from .gauges import ColdCathodeGauge, HotCathodeGauge, RoughGauge
from .others import RGA
from .pumps import GetterPump, IonPump, ScrollPump, TurboPump
from .valves import (ApertureValve, ControlOnlyValveNC, ControlOnlyValveNO,
                     ControlValve, FastShutter, NeedleValve, PneumaticValve,
                     PneumaticValveDA, PneumaticValveNO, ProportionalValve,
                     RightAngleManualValve)



import os.path

from pydm import Display


class BasicTable(Display):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ui.example_table.add_filter(
            'Hide Negative Values',
            self.neg_filter,
            active=True,
        )
        self.ui.example_table.add_filter(
            'Hide Even Values',
            self.even_filter,
            active=False,
        )
        self.ui.example_table.add_filter(
            'Hide Rows with 4 Character Names',
            self.four_filter,
            active=False,
        )

    def neg_filter(self, value_dict):
        return value_dict['readback'] >= 0

    def even_filter(self, value_dict):
        return value_dict['readback'] % 2

    def four_filter(self, value_dict):
        return len(value_dict['row_name']) != 4

    def ui_filename(self):
        return os.path.join(os.path.dirname(__file__), 'basic_table.ui')



from pydm.widgets import PyDMByteIndicator, PyDMChannel
from pydm.widgets.display_format import DisplayFormat, parse_value_for_display
from qtpy.QtCore import Property, Qt
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QTabWidget, QVBoxLayout, QWidget


class ByteIndicator_NegativeNums(PyDMByteIndicator):
    """
    Modified Byte Indicator Class.

    Overrides update_indicators function to allow for negative numbers to work
    for PyDmByteIndicator.
    """

    def update_indicators(self):
        if self._shift < 0:
            value = int(self.value) << abs(self._shift)
        else:
            value = int(self.value) >> self._shift

        bits = [(value >> i) & 1
                for i in range(self._num_bits)]
        for bit, indicator in zip(bits, self._indicators):
            if self._connected:
                if self._alarm_state == 3:
                    c = self._invalid_color
                else:
                    c = self._on_color if bit else self._off_color
            else:
                c = self._disconnected_color
            indicator.setColor(c)


class EPSByteIndicator(QWidget):
    """
    Widget for displaying EPS interlocks
    """

    _qt_designer_ = {
        "group": "PCDS Utilities",
        "is_container": False,
    }

    template_widget: ByteIndicator_NegativeNums

    _channels_prefix = None
    _value_channel = None
    _label_channel = None

    _value_pv = ""
    _label_pv = ""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.template_widget = ByteIndicator_NegativeNums()
        layout = QVBoxLayout()
        layout.addWidget(self.template_widget)
        self.setLayout(layout)

    def value_change(self, new_value):
        """
        Callback function when value changes
        """
        self.template_widget.value = new_value
        self.template_widget.update_indicators()

    def label_change(self, new_labels):
        """
        Callback function when the lables change
        """

        labels = parse_value_for_display(value=new_labels, precision=0,
                                         display_format_type=DisplayFormat.String)

        labels = labels.split(';')
        self.template_widget.numBits = len(labels)
        self.template_widget.labels = labels
        self.template_widget.update_indicators()

    def value_channel(self, connection):
        self.template_widget._connected = connection

    @Property(str)
    def channel(self):
        """
        PV of Base EPS strcture
        """
        return self._channels_prefix

    @channel.setter
    def channel(self, ch):
        """
        Set PV of Base EPS strcture
        """
        if ch != self._channels_prefix:
            self._value_pv = ch + ":nFlags_RBV"
            self._label_pv = ch + ":sFlagDesc_RBV"

            _value_channel = PyDMChannel(address=self._value_pv,
                                         connection_slot=self.value_channel,
                                         value_slot=self.value_change)

            _label_channel = PyDMChannel(address=self._label_pv,
                                         value_slot=self.label_change)
            _value_channel.connect()
            _label_channel.connect()

            self._channels_prefix = ch

    @Property(bool)
    def circles(self):
        return self.template_widget.circles

    @circles.setter
    def circles(self, circles):
        self.template_widget.circles = circles

    @Property(QTabWidget.TabPosition)
    def label_position(self):
        return self.template_widget._label_position

    @label_position.setter
    def label_position(self, position):
        self.template_widget._label_position = position
        self.template_widget.rebuild_layout()

    @Property(Qt.Orientation)
    def orientation(self):
        return self.template_widget._orientation

    @orientation.setter
    def orientation(self, orientation):
        self.template_widget._orientation = orientation
        self.template_widget.set_spacing()
        self.template_widget.rebuild_layout()

    @Property(QColor)
    def OnColor(self):
        return self.template_widget._on_color

    @OnColor.setter
    def OnColor(self, new_color):
        if self.template_widget._on_color != new_color:
            self.template_widget._on_color = new_color
            self.template_widget.update_indicators()

    @Property(QColor)
    def OffColor(self):
        return self.template_widget._off_color

    @OffColor.setter
    def OffColor(self, new_color):
        if self.template_widget._off_color != new_color:
            self.template_widget._off_color = new_color
            self.template_widget.update_indicators()



import logging
import os
from functools import partial

from pydm.widgets.channel import PyDMChannel
from pydm.widgets.enum_button import PyDMEnumButton
from pydm.widgets.label import PyDMLabel
from pydm.widgets.pushbutton import PyDMPushButton
from qtpy.QtCore import Property, Qt
from qtpy.QtWidgets import QGridLayout, QVBoxLayout

logger = logging.getLogger(__name__)


class InterlockMixin:
    """
    The InterlockMixin class adds the interlock channel and `interlocked`
    property to the widget.

    The interlocked property can be used at stylesheet in the following manner:


    .. code-block:: css

        *[interlocked="true"] {
            background-color: red;
        }

    Parameters
    ----------
    interlock_suffix : str
        The suffix to be used along with the channelPrefix from PCDSSymbolBase
        to compose the interlock channel address.
    """
    def __init__(self, interlock_suffix, **kwargs):
        self._interlock_suffix = interlock_suffix
        self._interlocked = False
        self._interlock_connected = False
        self.interlock_channel = None
        super().__init__(**kwargs)

    @Property(bool, designable=False)
    def interlocked(self):
        """
        Property used to query interlock state.

        Returns
        -------
        bool
        """
        return self._interlocked

    def create_channels(self):
        """
        This method invokes `create_channels` from the super classes and adds
        the `interlock_channel` to the widget along with a reset for the
        interlocked and interlock_connected variables.
        """
        super().create_channels()
        if not self._interlock_suffix:
            return

        self._interlocked = True
        self._interlock_connected = False

        self.interlock_channel = PyDMChannel(
            address="{}{}".format(self._channels_prefix,
                                  self._interlock_suffix),
            connection_slot=self.interlock_connection_changed,
            value_slot=self.interlock_value_changed
        )
        self.interlock_channel.connect()

    def status_tooltip(self):
        """
        This method adds the contribution of the interlock mixin into the
        general status tooltip.

        Returns
        -------
        str
        """
        status = super().status_tooltip()
        if status:
            status += os.linesep
        status += f"Interlocked: {self.interlocked}"
        return status

    def interlock_connection_changed(self, conn):
        """
        Callback invoked when the connection status changes for the Interlock
        Channel.

        Parameters
        ----------
        conn : bool
            True if connected, False otherwise.
        """
        self._interlock_connected = conn

    def interlock_value_changed(self, value):
        """
        Callback invoked when the value changes for the Interlock Channel.

        Parameters
        ----------
        value : int
            The value from the channel will be either 0 or 1 with 0 meaning
            that the widget is interlocked.
        """
        self._interlocked = value == 0
        self.controls_frame.setEnabled(not self._interlocked)
        self.update_stylesheet()
        self.update_status_tooltip()


class ErrorMixin:
    """
    The ErrorMixin class adds the error channel and `error`
    property to the widget.

    The error property can be used at stylesheet in the following manner:

    .. code-block:: css

        *[error="INVALID"] {
            color: purple;
        }

    Parameters
    ----------
    error_suffix : str
        The suffix to be used along with the channelPrefix from PCDSSymbolBase
        to compose the error channel address.
    """
    def __init__(self, error_suffix, **kwargs):
        self._error_suffix = error_suffix
        self._error = ""
        self._error_value = None
        self._error_enum = []
        self._error_connected = False
        self.error_channel = None
        super().__init__(**kwargs)

    @Property(str, designable=False)
    def error(self):
        """
        Property used to query the error state.

        Returns
        -------
        str
        """
        return self._error

    def create_channels(self):
        """
        This method invokes `create_channels` from the super classes and adds
        the `error_channel` to the widget along with a reset for the error and
        error_connected variables.
        """
        super().create_channels()
        if not self._error_suffix:
            return

        self._error_connected = False
        self._error = ""

        self.error_channel = PyDMChannel(
            address=f"{self._channels_prefix}{self._error_suffix}",
            connection_slot=self.error_connection_changed,
            value_slot=self.error_value_changed,
            enum_strings_slot=self.error_enum_changed
        )
        self.error_channel.connect()

    def status_tooltip(self):
        """
        This method adds the contribution of the error mixin into the general
        status tooltip.

        Returns
        -------
        str
        """
        status = super().status_tooltip()
        status += os.linesep
        status += f"Error: {self.error}"
        return status

    def error_connection_changed(self, conn):
        """
        Callback invoked when the connection status changes for the Error
        Channel.

        Parameters
        ----------
        conn : bool
            True if connected, False otherwise.
        """
        self._error_connected = conn

    def error_enum_changed(self, items):
        """
        Callback invoked when the enumeration strings change for the Error
        Channel.
        This callback triggers the update of the error message and also a
        repaint of the widget with the new stylesheet guidelines for the
        current error value.

        Parameters
        ----------
        items : tuple
            The string items
        """
        if items is None:
            return
        self._error_enum = items
        self._update_error_msg()

    def error_value_changed(self, value):
        """
        Callback invoked when the value change for the Error Channel.
        This callback triggers the update of the error message and also a
        repaint of the widget with the new stylesheet guidelines for the
        current error value.

        Parameters
        ----------
        value : int
        """
        if value is None:
            return
        self._error_value = value
        self._update_error_msg()

    def _update_error_msg(self):
        """
        Internal method that updates the error property and triggers an update
        on the stylesheet and tooltip.
        """
        if self._error_value is None:
            return
        if len(self._error_enum) > 0:
            try:
                self._error = self._error_enum[self._error_value]
            except IndexError:
                self._error = ""
        else:
            self._error = str(self._error_value)
        self.update_stylesheet()
        self.update_status_tooltip()


class StateMixin:
    """
    The StateMixin class adds the state channel and `state` property to the
    widget.

    The state property can be used at stylesheet in the following manner:

    .. code-block:: css

        *[state="Vented"] {
            border: 5px solid blue;
        }

    Parameters
    ----------
    state_suffix : str
        The suffix to be used along with the channelPrefix from PCDSSymbolBase
        to compose the state channel address.
    """
    def __init__(self, state_suffix, **kwargs):
        self._state_suffix = state_suffix
        self._state = ""
        self._state_value = None
        self._state_enum = []
        self._state_connected = False
        self.state_channel = None
        super().__init__(**kwargs)

    @Property(str, designable=False)
    def state(self):
        """
        Property used to query the state of the widget.

        Returns
        -------
        str
        """
        return self._state

    def create_channels(self):
        """
        This method invokes `create_channels` from the super classes and adds
        the `state_channel` to the widget along with a reset for the
        state and state_connected variables.
        """
        super().create_channels()
        if not self._state_suffix:
            return

        self._state_connected = False
        self._state = ""

        self.state_channel = PyDMChannel(
            address=f"{self._channels_prefix}{self._state_suffix}",
            connection_slot=self.state_connection_changed,
            value_slot=self.state_value_changed,
            enum_strings_slot=self.state_enum_changed
        )
        self.state_channel.connect()

    def status_tooltip(self):
        """
        This method adds the contribution of the state mixin into the general
        status tooltip.

        Returns
        -------
        str
        """
        status = super().status_tooltip()
        if status:
            status += os.linesep
        status += f"State: {self.state}"
        return status

    def state_connection_changed(self, conn):
        """
        Callback invoked when the connection status changes for the State
        Channel.

        Parameters
        ----------
        conn : bool
            True if connected, False otherwise.
        """
        self._state_connected = conn

    def state_enum_changed(self, items):
        """
        Callback invoked when the enumeration strings change for the State
        Channel.
        This callback triggers the update of the state message and also a
        repaint of the widget with the new stylesheet guidelines for the
        current state value.

        Parameters
        ----------
        items : tuple
            The string items
        """
        if items is None:
            return
        self._state_enum = items
        self._update_state_msg()

    def state_value_changed(self, value):
        """
        Callback invoked when the value change for the State Channel.
        This callback triggers the update of the state message and also a
        repaint of the widget with the new stylesheet guidelines for the
        current state value.

        Parameters
        ----------
        value : int
        """
        if value is None:
            return
        self._state_value = value
        self._update_state_msg()

    def _update_state_msg(self):
        """
        Internal method that updates the state property and triggers an update
        on the stylesheet and tooltip.
        """
        if self._state_value is None:
            return
        if len(self._state_enum) > 0:
            try:
                self._state = self._state_enum[self._state_value]
            except IndexError:
                self._state = ""
        else:
            self._state = str(self._state_value)
        self.update_stylesheet()
        self.update_status_tooltip()


class OpenCloseStateMixin:
    """
    The OpenCloseStateMixin class adds two channels (Open and Close State) and
    a `state` property based on a combination of the two channels to the
    widget.

    The state property can be used at stylesheet in the following manner:

    .. code-block:: css

        *[state="Close"] {
            background-color: blue;
        }

    Parameters
    ----------
    open_suffix : str
        The suffix to be used along with the channelPrefix from PCDSSymbolBase
        to compose the open state channel address.
    close_suffix : str
        The suffix to be used along with the channelPrefix from PCDSSymbolBase
        to compose the close state channel address.
    """
    def __init__(self, open_suffix, close_suffix, **kwargs):
        self._open_suffix = open_suffix
        self._close_suffix = close_suffix

        self._state_open = False
        self._state_close = False

        self._open_connected = False
        self._close_connected = False

        self.state_open_channel = None
        self.state_close_channel = None
        super().__init__(**kwargs)

    @Property(str, designable=False)
    def state(self):
        """
        Property used to query the state of the widget.

        Returns
        -------
        str
            The return string will either be `Open`, `Close` or `INVALID` when
            it was not possible to determine the state.
        """
        if self._state_open == self._state_close:
            return "INVALID"
        if self._state_open:
            return "Open"
        else:
            return "Close"

    def create_channels(self):
        """
        This method invokes `create_channels` from the super classes and adds
        the `state_open_channel` and `state_close_channel` to the widget along
        with a reset for the state and interlock_connected variables.
        """
        super().create_channels()
        if not self._open_suffix or not self._close_suffix:
            return

        self._open_connected = False
        self._close_connected = False
        self._state_open = False
        self._state_close = False
        self._state = "INVALID"

        self.state_open_channel = PyDMChannel(
            address=f"{self._channels_prefix}{self._open_suffix}",
            connection_slot=partial(self.state_connection_changed, "OPEN"),
            value_slot=partial(self.state_value_changed, "OPEN")
        )
        self.state_open_channel.connect()

        self.state_close_channel = PyDMChannel(
            address=f"{self._channels_prefix}{self._close_suffix}",
            connection_slot=partial(self.state_connection_changed, "CLOSE"),
            value_slot=partial(self.state_value_changed, "CLOSE")
        )
        self.state_close_channel.connect()

    def status_tooltip(self):
        """
        This method adds the contribution of the open close state mixin into
        the general status tooltip.

        Returns
        -------
        str
        """
        status = super().status_tooltip()
        if status:
            status += os.linesep
        status += f"State: {self.state}"
        return status

    def state_connection_changed(self, which, conn):
        """
        Callback invoked when the connection status changes for one of the
        channels in this mixin.

        Parameters
        ----------
        which : str
            String defining which channel is sending the information. It must
            be either "OPEN" or "CLOSE".
        conn : bool
            True if connected, False otherwise.
        """
        if which == "OPEN":
            self._open_connected = conn
        else:
            self._close_connected = conn

    def state_value_changed(self, which, value):
        """
        Callback invoked when the value changes for one of the channels in this
        mixin.

        Parameters
        ----------
        which : str
            String defining which channel is sending the information. It must
            be either "OPEN" or "CLOSE".
        value : int
            The value from the channel which will be either 0 or 1 with 1
            meaning that a certain state is active.
        """
        if which == "OPEN":
            self._state_open = value
        else:
            self._state_close = value

        self.update_stylesheet()
        self.update_status_tooltip()


class ButtonControl:
    """
    The ButtonControl class adds a PyDMEnumButton to the widget for controls.

    Parameters
    ----------
    command_suffix : str
        The suffix to be used along with the channelPrefix from PCDSSymbolBase
        to compose the command button channel address.
    """
    def __init__(self, command_suffix, **kwargs):
        self._command_suffix = command_suffix
        self._orientation = Qt.Horizontal
        self.control_btn = PyDMEnumButton()
        self.control_btn.checkable = False
        self.controlButtonHorizontal = True
        self.controls_layout = QVBoxLayout()
        self.controls_layout.setSpacing(2)
        self.controls_layout.setContentsMargins(0, 10, 0, 0)
        super().__init__(**kwargs)
        self.controls_frame.setLayout(self.controls_layout)
        self.controls_frame.layout().addWidget(self.control_btn)

    @Property(bool)
    def controlButtonHorizontal(self):
        return self._orientation == Qt.Horizontal

    @controlButtonHorizontal.setter
    def controlButtonHorizontal(self, checked):
        if checked:
            self._orientation = Qt.Horizontal
            self.control_btn.setMinimumSize(100, 40)
        else:
            self._orientation = Qt.Vertical
            self.control_btn.setMinimumSize(40, 80)

        self.control_btn.orientation = self._orientation

    def create_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        created.
        This method also sets the channel address for the control button.
        """
        super().create_channels()
        if self._channels_prefix:
            self.control_btn.channel = "{}{}".format(self._channels_prefix,
                                                     self._command_suffix)

    def destroy_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        destroyed.
        This method also clears the channel address for the control button.
        """
        super().destroy_channels()
        self.control_btn.channel = None


class LabelControl:
    """
    The LabelControl class adds a PyDMLabel to the widget for controls.

    Parameters
    ----------
    readback_suffix : str
        The suffix to be used along with the channelPrefix from PCDSSymbolBase
        to compose the readback label channel address.
    readback_name : str
        The name to be set to the PyDMLabel so one can refer to it by name
        with stylesheet
    """
    def __init__(self, readback_suffix, readback_name,
                 **kwargs):
        self._readback_suffix = readback_suffix
        self.readback_label = PyDMLabel()
        if readback_name:
            self.readback_label.setObjectName(readback_name)
        self.readback_label.setAlignment(Qt.AlignCenter)
        self.controls_layout = QVBoxLayout()
        self.controls_layout.setSpacing(2)
        self.controls_layout.setContentsMargins(0, 10, 0, 0)
        super().__init__(**kwargs)
        self.controls_frame.setLayout(self.controls_layout)
        self.controls_frame.layout().addWidget(self.readback_label)

    def create_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        created.
        This method also sets the channel address for the control button.
        """
        super().create_channels()
        if self._channels_prefix:
            self.readback_label.channel = "{}{}".format(self._channels_prefix,
                                                        self._readback_suffix)

    def destroy_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        destroyed.
        This method also clears the channel address for the control button.
        """
        super().destroy_channels()
        self.readback_label.channel = None


class ButtonLabelControl(ButtonControl):
    """
    The ButtonLabelControl class adds a PyDMEnumButton and a PyDMLabel to the
    widget for controls.

    Parameters
    ----------
    command_suffix : str
        The suffix to be used along with the channelPrefix from PCDSSymbolBase
        to compose the command button channel address.
    readback_suffix : str
        The suffix to be used along with the channelPrefix from PCDSSymbolBase
        to compose the readback label channel address.
    readback_name : str
        The name to be set to the PyDMLabel so one can refer to it by name
        with stylesheet
    """
    def __init__(self, command_suffix, readback_suffix, readback_name,
                 **kwargs):
        self._readback_suffix = readback_suffix

        self.readback_label = PyDMLabel()
        self.readback_label.setObjectName(readback_name)
        self.readback_label.setAlignment(Qt.AlignCenter)
        super().__init__(command_suffix, **kwargs)
        self.controls_frame.layout().insertWidget(0, self.readback_label)

    def create_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        created.
        This method also sets the channel address for the control button.
        """
        super().create_channels()
        if self._channels_prefix:
            self.readback_label.channel = "{}{}".format(self._channels_prefix,
                                                        self._readback_suffix)

    def destroy_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        destroyed.
        This method also clears the channel address for the control button.
        """
        super().destroy_channels()
        self.readback_label.channel = None


class MultipleButtonControl:
    """
    The MultipleButtonControl class adds multiple PyDMPushButton instances to
    the widget for controls.

    Parameters
    ----------
    commands : list
        List of dictionaries containing the specifications for the buttons.
        Required keys for now are:

        - suffix: str
            suffix to be used along with the channelPrefix from PCDSSymbolBase
            to compose the command button channel address
        - text: str
            the text to display at the button
        - value
            the value to be written when the button is pressed
    """
    def __init__(self, *, commands, **kwargs):
        self._command_buttons_config = commands
        self._orientation = Qt.Horizontal
        self.buttons = []
        self.create_buttons()

        super().__init__(**kwargs)
        self.controls_frame.setLayout(QGridLayout())
        self.controlButtonHorizontal = True

    @Property(bool)
    def controlButtonHorizontal(self):
        return self._orientation == Qt.Horizontal

    @controlButtonHorizontal.setter
    def controlButtonHorizontal(self, checked):
        self.clear_control_layout()

        self._orientation = Qt.Vertical
        if checked:
            self._orientation = Qt.Horizontal

        layout = self.controls_frame.layout()

        if self._orientation == Qt.Vertical:
            for i, btn in enumerate(self.buttons):
                layout.addWidget(btn, i, 0)
        elif self._orientation == Qt.Horizontal:
            for i, btn in enumerate(self.buttons):
                layout.addWidget(btn, 0, i)

    def clear_control_layout(self):
        """
        Remove all inner widgets from the control layout
        """
        layout = self.controls_frame.layout()
        if not isinstance(layout, QGridLayout):
            return
        for col in range(0, layout.columnCount()):
            for row in range(0, layout.rowCount()):
                item = layout.itemAtPosition(row, col)
                if item is not None:
                    w = item.widget()
                    if w is not None:
                        layout.removeWidget(w)

    def create_buttons(self):
        for btn in self._command_buttons_config:
            try:
                text = btn['text']
                value = btn['value']
                btn = PyDMPushButton(label=text, pressValue=value)
                self.buttons.append(btn)
            except KeyError:
                logger.exception('Invalid config for MultipleButtonControl.')

    def create_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        created.
        This method also sets the channel address for the control button.
        """
        super().create_channels()
        if self._channels_prefix:
            for idx, btn in enumerate(self.buttons):
                suffix = self._command_buttons_config[idx]['suffix']
                btn.channel = f"{self._channels_prefix}{suffix}"

    def destroy_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        destroyed.
        This method also clears the channel address for the control button.
        """
        super().destroy_channels()
        for btn in self.buttons:
            btn.channel = None



from pydm.widgets.channel import PyDMChannel
from pydm.widgets.pushbutton import PyDMPushButton
from qtpy.QtCore import Property, QSize, Qt
from qtpy.QtWidgets import QGridLayout

from ..icons.valves import (ApertureValveSymbolIcon,
                            ControlOnlyValveSymbolIcon, ControlValveSymbolIcon,
                            FastShutterSymbolIcon, NeedleValveSymbolIcon,
                            PneumaticValveDASymbolIcon,
                            PneumaticValveNOSymbolIcon,
                            PneumaticValveSymbolIcon,
                            ProportionalValveSymbolIcon,
                            RightAngleManualValveSymbolIcon)
from .base import ContentLocation, PCDSSymbolBase
from .mixins import (ButtonControl, ErrorMixin, InterlockMixin,
                     MultipleButtonControl, StateMixin)


class PneumaticValve(
    InterlockMixin, ErrorMixin, StateMixin, ButtonControl, PCDSSymbolBase
):
    """
    A Symbol Widget representing a Pneumatic Valve with the proper icon and
    controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------+
    |Property   |Values                                                 |
    +===========+=======================================================+
    |interlocked|`true` or `false`                                      |
    +-----------+-------------------------------------------------------+
    |error      |`Vented`, `At Vacuum`, `Differential Pressure` or      |
    |           |`Lost Vacuum`                                          |
    +-----------+-------------------------------------------------------+
    |state      |`Open`, `Closed`, `Moving`, `Invalid`                  |
    +-----------+-------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        PneumaticValve[interlocked="true"] #interlock {
            border: 5px solid red;
        }
        PneumaticValve[interlocked="false"] #interlock {
            border: 0px;
        }
        PneumaticValve[interlocked="true"] #icon {
            qproperty-interlockBrush: #FF0000;
        }
        PneumaticValve[interlocked="false"] #icon {
            qproperty-interlockBrush: #00FF00;
        }
        PneumaticValve[error="Lost Vacuum"] #icon {
            qproperty-penStyle: "Qt::DotLine";
            qproperty-penWidth: 2;
            qproperty-brush: red;
        }
        PneumaticValve[state="Open"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }

    _interlock_suffix = ":OPN_OK_RBV"
    _error_suffix = ":STATE_RBV"
    _state_suffix = ":POS_STATE_RBV"
    _command_suffix = ":OPN_SW"

    NAME = "Pneumatic Valve"
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VGC"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            error_suffix=self._error_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            **kwargs)
        self.icon = PneumaticValveSymbolIcon(parent=self)

    def sizeHint(self):
        return QSize(180, 70)


class ApertureValve(
    InterlockMixin, ErrorMixin, StateMixin, ButtonControl, PCDSSymbolBase
):
    """
    A Symbol Widget representing an Aperture Valve with the proper icon and
    controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------+
    |Property   |Values                                                 |
    +===========+=======================================================+
    |interlocked|`true` or `false`                                      |
    +-----------+-------------------------------------------------------+
    |error      |`Vented`, `At Vacuum`, `Differential Pressure` or      |
    |           |`Lost Vacuum`                                          |
    +-----------+-------------------------------------------------------+
    |state      |`Open`, `Close`, `Moving` or `INVALID`                 |
    +-----------+-------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        ApertureValve [interlocked="true"] #interlock {
            border: 5px solid red;
        }
        ApertureValve [interlocked="false"] #interlock {
            border: 0px;
        }
        ApertureValve [interlocked="true"] #icon {
            qproperty-interlockBrush: #FF0000;
        }
        ApertureValve [interlocked="false"] #icon {
            qproperty-interlockBrush: #00FF00;
        }
        ApertureValve [error="Lost Vacuum"] #icon {
            qproperty-penStyle: "Qt::DotLine";
            qproperty-penWidth: 2;
            qproperty-brush: red;
        }
        ApertureValve [state="Open"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    _interlock_suffix = ":OPN_OK_RBV"
    _error_suffix = ":STATE_RBV"
    _state_suffix = ":POS_STATE_RBV"
    _command_suffix = ":OPN_SW"

    NAME = "Aperture Valve"
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VRC"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            error_suffix=self._error_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            **kwargs)
        self.icon = ApertureValveSymbolIcon(parent=self)

    def sizeHint(self):
        return QSize(180, 70)


class FastShutter(
    InterlockMixin, ErrorMixin, StateMixin, MultipleButtonControl, PCDSSymbolBase
):
    """
    A Symbol Widget representing a Fast Shutter with the proper icon and
    controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------------+
    |Property   |Values                                                       |
    +===========+=============================================================+
    |interlocked|`true` or `false`                                            |
    +-----------+-------------------------------------------------------------+
    |error      |`true`, or `false`                                           |
    +-----------+-------------------------------------------------------------+
    |state      |`Open`, `Close` `Moving` or `INVALID`                        |
    +-----------+-------------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        FastShutter [interlocked="true"] #interlock {
            border: 5px solid red;
        }
        FastShutter [interlocked="false"] #interlock {
            border: 0px;
        }
        FastShutter [error="true"] #icon {
            qproperty-penStyle: "Qt::DotLine";
            qproperty-penWidth: 2;
            qproperty-brush: red;
            qproperty-arrowBrush: #00FF00;
        }
        FastShutter [state="Open"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    _interlock_suffix = ":OPN_OK_RBV"
    _error_suffix = ":STATE_RBV"
    _state_suffix = ":POS_STATE_RBV"
    _command_buttons = [
        {"suffix": ":OPN_SW", "text": "OPEN", "value": 1},
        {"suffix": ":CLS_SW", "text": "CLOSE", "value": 1},
    ]

    NAME = "Fast Shutter"
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VFS"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            error_suffix=self._error_suffix,
            state_suffix=self._state_suffix,
            commands=self._command_buttons,
            **kwargs)
        self.icon = FastShutterSymbolIcon(parent=self)

    def sizeHint(self):
        return QSize(180, 70)


class NeedleValve(InterlockMixin, StateMixin, ButtonControl, PCDSSymbolBase):
    """
    A Symbol Widget representing a Needle Valve with the proper icon and
    controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------------+
    |Property   |Values                                                       |
    +===========+=============================================================+
    |interlocked|`true` or `false`                                            |
    +-----------+-------------------------------------------------------------+
    |state      |`Close`, `Open`, `PressureControl`, `ManualControl`          |
    +-----------+-------------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        NeedleValve [interlocked="true"] #interlock {
            border: 5px solid red;
        }
        NeedleValve [interlocked="false"] #interlock {
            border: 0px;
        }
        FastShutter [state="Open"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    _interlock_suffix = ":ILK_OK_RBV"
    _state_suffix = ":STATE_RBV"
    _command_suffix = ":OPN_SW"

    NAME = "Needle Valve"
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VCN"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            **kwargs)
        self.icon = NeedleValveSymbolIcon(parent=self)

    def sizeHint(self):
        return QSize(180, 70)


class ProportionalValve(InterlockMixin, StateMixin, ButtonControl, PCDSSymbolBase):
    """
    A Symbol Widget representing a Proportional Valve with the proper icon and
    controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------------+
    |Property   |Values                                                       |
    +===========+=============================================================+
    |interlocked|`true` or `false`                                            |
    +-----------+-------------------------------------------------------------+
    |state      |`Close`, `Open`, `PressureControl`, `ManualControl`          |
    +-----------+-------------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        ProportionalValve [interlocked="true"] #interlock {
            border: 5px solid red;
        }
        ProportionalValve [interlocked="false"] #interlock {
            border: 0px;
        }
        ProportionalValve [state="Open"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    _interlock_suffix = ":ILK_OK_RBV"
    _state_suffix = ":STATE_RBV"
    _command_suffix = ":OPN_SW"

    NAME = "Proportional Valve"
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VRC"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            **kwargs)
        self.icon = ProportionalValveSymbolIcon(parent=self)

    def sizeHint(self):
        return QSize(180, 70)


class RightAngleManualValve(PCDSSymbolBase):
    """
    A Symbol Widget representing a Right Angle Manual Valve with the proper
    icon.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    NAME = "Right Angle Manual Valve"

    def __init__(self, parent=None, **kwargs):
        self._controls_location = ContentLocation.Hidden
        super().__init__(parent=parent, **kwargs)
        self.icon = RightAngleManualValveSymbolIcon(parent=self)

    def sizeHint(self):
        """
        Suggested initial size for the widget.

        Returns
        -------
        size : QSize
        """
        return QSize(40, 40)

    @Property(str, designable=False)
    def channelsPrefix(self):
        return super().channelsPrefix

    @Property(bool, designable=False)
    def showIcon(self):
        return super().showIcon

    @Property(ContentLocation, designable=False)
    def controlsLocation(self):
        return super().controlsLocation


class ControlValve(
    InterlockMixin, ErrorMixin, StateMixin, ButtonControl, PCDSSymbolBase
):
    """
    A Symbol Widget representing a Control Valve with the proper icon and
    controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------+
    |Property   |Values                                                 |
    +===========+=======================================================+
    |interlocked|`true` or `false`                                      |
    +-----------+-------------------------------------------------------+
    |error      |`Vented`, `At Vacuum`, `Differential Pressure` or      |
    |           |`Lost Vacuum`                                          |
    +-----------+-------------------------------------------------------+
    |state      |`Open`, `Close` `Moving` or `INVALID`                  |
    +-----------+-------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        ControlValve [interlocked="true"] #interlock {
            border: 5px solid red;
        }
        ControlValve [interlocked="false"] #interlock {
            border: 0px;
        }
        ControlValve [interlocked="true"] #icon {
            qproperty-interlockBrush: #FF0000;
        }
        ControlValve [interlocked="false"] #icon {
            qproperty-interlockBrush: #00FF00;
        }
        ControlValve [error="Lost Vacuum"] #icon {
            qproperty-penStyle: "Qt::DotLine";
            qproperty-penWidth: 2;
            qproperty-brush: red;
        }
        ControlValve [state="Open"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    NAME = 'Control Valve with Readback'
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VVC"

    _interlock_suffix = ":OPN_OK_RBV"
    _error_suffix = ":STATE_RBV"
    _state_suffix = ":POS_STATE_RBV"
    _command_suffix = ":OPN_SW"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            error_suffix=self._error_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            **kwargs)
        self.icon = ControlValveSymbolIcon(parent=self)

    def sizeHint(self):
        return QSize(180, 70)


class ControlOnlyValveNC(InterlockMixin, StateMixin, ButtonControl, PCDSSymbolBase):
    """
    A Symbol Widget representing a Normally Closed Control Valve with the
    proper icon and controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------+
    |Property   |Values                                                 |
    +===========+=======================================================+
    |interlocked|`true` or `false`                                      |
    +-----------+-------------------------------------------------------+
    |state      |`Open`, `Close` or `INVALID`                           |
    +-----------+-------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        ControlOnlyValveNC [interlocked="true"] #interlock {
            border: 5px solid red;
        }
        ControlOnlyValveNC [interlocked="false"] #interlock {
            border: 0px;
        }
        ControlOnlyValveNC [interlocked="true"] #icon {
            qproperty-interlockBrush: #FF0000;
        }
        ControlOnlyValveNC [interlocked="false"] #icon {
            qproperty-interlockBrush: #00FF00;
        }
        ControlOnlyValveNC [error="Lost Vacuum"] #icon {
            qproperty-penStyle: "Qt::DotLine";
            qproperty-penWidth: 2;
            qproperty-brush: red;
        }
        ControlOnlyValveNC [state="Open"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    NAME = 'Normally Closed Control Valve with No Readback'
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VVC"

    _interlock_suffix = ":OPN_OK_RBV"
    _state_suffix = ':OPN_DO_RBV'
    _command_suffix = ":OPN_SW"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            **kwargs)
        self.icon = ControlOnlyValveSymbolIcon(parent=self)

    def sizeHint(self):
        return QSize(180, 70)


class ControlOnlyValveNO(InterlockMixin, StateMixin, ButtonControl, PCDSSymbolBase):
    """
    A Symbol Widget representing a Normally Open Control Valve with the
    proper icon and controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------+
    |Property   |Values                                                 |
    +===========+=======================================================+
    |interlocked|`true` or `false`                                      |
    +-----------+-------------------------------------------------------+
    |state      |`Open`, `Close` or `INVALID`                           |
    +-----------+-------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        ControlOnlyValveNO [interlocked="true"] #interlock {
            border: 5px solid red;
        }
        ControlOnlyValveNO [interlocked="false"] #interlock {
            border: 0px;
        }
        ControlOnlyValveNO [interlocked="true"] #icon {
            qproperty-interlockBrush: #FF0000;
        }
        ControlOnlyValveNO [interlocked="false"] #icon {
            qproperty-interlockBrush: #00FF00;
        }
        ControlOnlyValveNO [error="Lost Vacuum"] #icon {
            qproperty-penStyle: "Qt::DotLine";
            qproperty-penWidth: 2;
            qproperty-brush: red;
        }
        ControlOnlyValveNO [state="Open"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    NAME = 'Normally Open Control Valve with No Readback'
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VVCNO"

    _interlock_suffix = ":CLS_OK_RBV"
    _state_suffix = ':CLS_DO_RBV'
    _command_suffix = ":CLS_SW"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            **kwargs)
        self.icon = ControlOnlyValveSymbolIcon(parent=self)


class PneumaticValveNO(
    InterlockMixin, ErrorMixin, StateMixin, ButtonControl, PCDSSymbolBase
):
    """
    A Symbol Widget representing a Normally Open Pneumatic Valve with the
    proper icon and controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------+
    |Property   |Values                                                 |
    +===========+=======================================================+
    |interlocked|`true` or `false`                                      |
    +-----------+-------------------------------------------------------+
    |error      |`Vented`, `At Vacuum`, `Differential Pressure` or      |
    |           |`Lost Vacuum`                                          |
    +-----------+-------------------------------------------------------+
    |state      |`OPEN`, `CLOSED`, `MOVING`, `INVALID`                  |
    +-----------+-------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        PneumaticValveNO[interlocked="true"] #interlock {
            border: 5px solid red;
        }
        PneumaticValveNO[interlocked="false"] #interlock {
            border: 0px;
        }
        PneumaticValveNO[interlocked="true"] #icon {
            qproperty-interlockBrush: #FF0000;
        }
        PneumaticValveNO[interlocked="false"] #icon {
            qproperty-interlockBrush: #00FF00;
        }
        PneumaticValveNO[error="Lost Vacuum"] #icon {
            qproperty-penStyle: "Qt::DotLine";
            qproperty-penWidth: 2;
            qproperty-brush: red;
        }
        PneumaticValveNO[state="OPEN"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    _interlock_suffix = ":CLS_OK_RBV"
    _error_suffix = ":STATE_RBV"
    _state_suffix = ":POS_STATE_RBV"
    _command_suffix = ":CLS_SW"

    NAME = "Pneumatic Valve NO"
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VVCNO"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            error_suffix=self._error_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            **kwargs)
        self.icon = PneumaticValveNOSymbolIcon(parent=self)

    def sizeHint(self):
        return QSize(180, 70)


class PneumaticValveDA(InterlockMixin, ErrorMixin, StateMixin, PCDSSymbolBase):
    """
    A Symbol Widget representing a dual-acting Pneumatic Valve with
    the proper icon and controls.

    This needs to modify the normal interlock logic because it has
    two interlock PVs instead of one.

    This also needs to completely re-implement the button control
    logic because it needs to have separate PVs for opening and
    closing the valve, and the ButtonControl mixin assumes one
    PV with enum values.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------+
    |Property   |Values                                                 |
    +===========+=======================================================+
    |interlocked|`true` or `false`                                      |
    +-----------+-------------------------------------------------------+
    |error      |`Vented`, `At Vacuum`, `Differential Pressure` or      |
    |           |`Lost Vacuum`                                          |
    +-----------+-------------------------------------------------------+
    |state      |`Open`, `Closed`, `Moving`, `Invalid`                  |
    +-----------+-------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        PneumaticValveDA[interlocked="true"] #interlock {
            border: 5px solid red;
        }
        PneumaticValveDA[interlocked="false"] #interlock {
            border: 0px;
        }
        PneumaticValveDA[interlocked="true"] #icon {
            qproperty-interlockBrush: #FF0000;
        }
        PneumaticValveDA[interlocked="false"] #icon {
            qproperty-interlockBrush: #00FF00;
        }
        PneumaticValveDA[error="Lost Vacuum"] #icon {
            qproperty-penStyle: "Qt::DotLine";
            qproperty-penWidth: 2;
            qproperty-brush: red;
        }
        PneumaticValveDA[state="Open"] #icon {
            qproperty-penColor: green;
            qproperty-penWidth: 2;
        }
    """

    _qt_designer_ = {
        "group": "PCDS Valves",
        "is_container": False,
    }
    _interlock_suffix = ":OPN_OK_RBV"
    _cls_interlock_suffix = ":CLS_OK_RBV"
    _error_suffix = ":STATE_RBV"
    _state_suffix = ":POS_STATE_RBV"
    _open_command_suffix = ":OPN_SW"
    _close_command_suffix = ":CLS_SW"

    NAME = "Pneumatic Valve DA"
    EXPERT_OPHYD_CLASS = "pcdsdevices.valve.VRCDA"

    def __init__(self, parent=None, **kwargs):
        self._cls_interlocked = False
        self._cls_interlock_connected = False
        self.cls_interlock_channel = None
        self.controls_layout = None
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            error_suffix=self._error_suffix,
            state_suffix=self._state_suffix,
            **kwargs)
        self.icon = PneumaticValveDASymbolIcon(parent=self)
        self.open_btn = PyDMPushButton(
            label='OPEN',
            pressValue=1,
        )
        self.cls_btn = PyDMPushButton(
            label='CLOSE',
            pressValue=1,
        )
        self.open_btn.setFixedSize(55, 25)
        self.cls_btn.setFixedSize(55, 25)
        self.controls_layout = QGridLayout()
        self.controls_layout.setSpacing(6)
        self.controls_layout.setContentsMargins(5, 5, 5, 5)
        self.controls_frame.setLayout(self.controls_layout)
        self.controlButtonHorizontal = True

    @Property(bool, designable=False)
    def interlocked(self):
        """
        Property used to query interlock state.

        Returns
        -------
        bool
        """
        return self._interlocked or self._cls_interlocked

    @Property(bool)
    def controlButtonHorizontal(self):
        return self._orientation == Qt.Horizontal

    @controlButtonHorizontal.setter
    def controlButtonHorizontal(self, checked):
        if checked:
            self._orientation = Qt.Horizontal
        else:
            self._orientation = Qt.Vertical

        self.rearrange_button_layout()

    def rearrange_button_layout(self):
        if self._orientation == Qt.Horizontal:
            self.controls_frame.layout().addWidget(self.open_btn, 0, 1)
            self.controls_frame.layout().addWidget(self.cls_btn, 0, 0)
        else:
            self.controls_frame.layout().addWidget(self.cls_btn, 1, 0)
            self.controls_frame.layout().addWidget(self.open_btn, 0, 0)

    def create_channels(self):
        """
        Add a second interlock channel and the button channels.

        The second interlock channel is used to check if the closing
        action of the valve is permitted.

        The button channels allow us to open and close the valves.
        """
        super().create_channels()

        self._cls_interlocked = True
        self._cls_interlock_connected = False

        self.cls_interlock_channel = PyDMChannel(
            address="{}{}".format(self._channels_prefix,
                                  self._cls_interlock_suffix),
            connection_slot=self.cls_interlock_connection_changed,
            value_slot=self.cls_interlock_value_changed
        )
        self.cls_interlock_channel.connect()

        self.open_btn.channel = "{}{}".format(
            self._channels_prefix,
            self._open_command_suffix,
        )
        self.cls_btn.channel = "{}{}".format(
            self._channels_prefix,
            self._close_command_suffix,
        )

    def destroy_channels(self):
        """
        Method invoked when the channels associated with the widget must be
        destroyed.
        This method also clears the channel address for the control buttons
        and close interlock.
        """
        if self.cls_interlock_channel is not None:
            self.cls_interlock_channel.disconnect()
        self.cls_interlock_channel = None
        self.open_btn.channel = None
        self.cls_btn.channel = None

    def cls_interlock_connection_changed(self, conn):
        """
        Callback invoked when the connection status changes for the Interlock
        Channel.

        Neither this nor the open interlock connection state are currently
        used, but this was included for completeness.

        Parameters
        ----------
        conn : bool
            True if connected, False otherwise.
        """
        self._cls_interlock_connected = conn

    def interlock_value_changed(self, value):
        """
        Callback invoked when the value changes for the Interlock Channel.

        Parameters
        ----------
        value : int
            The value from the channel will be either 0 or 1 with 0 meaning
            that the widget is interlocked.
        """
        self._interlocked = value == 0
        self.open_btn.setEnabled(not self._interlocked)
        self.update_da_interlock()

    def cls_interlock_value_changed(self, value):
        """
        Callback invoked when the value changes for the Interlock Channel.

        Parameters
        ----------
        value : int
            The value from the channel will be either 0 or 1 with 0 meaning
            that the widget is interlocked.
        """
        self._cls_interlocked = value == 0
        self.cls_btn.setEnabled(not self._cls_interlocked)
        self.update_da_interlock()

    def update_da_interlock(self):
        """
        Update the double-acting interlock state when either pv changes.
        """
        self.update_stylesheet()
        self.update_status_tooltip()

    def sizeHint(self):
        return QSize(180, 70)



from qtpy.QtCore import Property, QSize

from ..icons.others import RGASymbolIcon
from .base import ContentLocation, PCDSSymbolBase


class RGA(PCDSSymbolBase):
    """
    A Symbol Widget representing a Residual Gas Analyzer with the proper icon.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+

    """

    _qt_designer_ = {
        "group": "PCDS Others",
        "is_container": False,
    }
    NAME = "Residual Gas Analyzer"

    def __init__(self, parent=None, **kwargs):
        self._controls_location = ContentLocation.Hidden
        super().__init__(parent=parent, **kwargs)
        self.icon = RGASymbolIcon(parent=self)

    def sizeHint(self):
        """
        Suggested initial size for the widget.

        Returns
        -------
        size : QSize
        """
        return QSize(40, 40)

    @Property(str, designable=False)
    def channelsPrefix(self):
        return super().channelsPrefix

    @Property(bool, designable=False)
    def showIcon(self):
        return super().showIcon

    @Property(ContentLocation, designable=False)
    def controlsLocation(self):
        return super().controlsLocation



import math

from qtpy.QtCore import Property, QPointF, QRectF
from qtpy.QtGui import QBrush, QColor, QPainterPath, QPolygonF, QTransform

from .base import BaseSymbolIcon


class ScrollPumpSymbolIcon(BaseSymbolIcon):
    """
    A widget with a scroll pump symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._center_brush = QBrush(QColor("transparent"))

    @Property(QBrush)
    def centerBrush(self):
        return self._center_brush

    @centerBrush.setter
    def centerBrush(self, new_brush):
        if new_brush != self._center_brush:
            self._center_brush = new_brush
            self.update()

    def draw_icon(self, painter):
        painter.drawEllipse(QPointF(0.5, 0.5), 0.5, 0.5)
        painter.drawChord(QRectF(0.0, 0.0, 1.0, 1.0), 45 * 16, -120 * 16)
        painter.drawChord(QRectF(0.0, 0.0, 1.0, 1.0), 135 * 16, 120 * 16)
        circle_arrow_point = QPointF(0.3, 0.5)

        brush = painter.brush()
        pen = painter.pen()

        painter.setBrush(self.centerBrush)
        painter.setPen(QColor("transparent"))
        painter.drawEllipse(QPointF(0.5, 0.5), 0.2, 0.2)

        painter.setBrush(brush)
        painter.setPen(pen)

        painter.drawArc(QRectF(0.3, 0.3, 0.4, 0.4), 90 * 16, -270 * 16)

        arrow = QPolygonF(
            [QPointF(-0.025, 0.0), QPointF(0.025, 0.0), QPointF(0.0, -0.025)])
        painter.setBrush(QBrush(QColor(0, 0, 0)))
        painter.drawPolygon(arrow.translated(circle_arrow_point))


class IonPumpSymbolIcon(BaseSymbolIcon):
    """
    A widget with an ion pump symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """

    def draw_icon(self, painter):
        painter.drawEllipse(QPointF(0.5, 0.5), 0.5, 0.5)
        painter.drawChord(QRectF(0.0, 0.0, 1.0, 1.0), 45 * 16, -120 * 16)
        painter.drawChord(QRectF(0.0, 0.0, 1.0, 1.0), 135 * 16, 120 * 16)
        bottom_arrow_point = QPointF(0.5, 0.8)
        painter.drawLine(bottom_arrow_point, QPointF(0.5, 0.7))
        curve_start = QPointF(0.5, 0.7)
        bend_angle = 25
        curve_end_l = QPointF(
            0.4 * math.cos(math.radians(90 + bend_angle)) + 0.5,
            -0.4 * math.sin(math.radians(90 + bend_angle)) + 0.5)
        c1 = QPointF(0.5, 0.4)
        path = QPainterPath(curve_start)
        path.quadTo(c1, curve_end_l)
        painter.drawPath(path)
        curve_end_r = QPointF(
            0.4 * math.cos(math.radians(90 - bend_angle)) + 0.5,
            -0.4 * math.sin(math.radians(90 - bend_angle)) + 0.5)
        path = QPainterPath(curve_start)
        path.quadTo(c1, curve_end_r)
        painter.drawPath(path)
        # Draw the arrow end-caps
        painter.setBrush(QBrush(QColor(0, 0, 0)))
        arrow = QPolygonF(
            [QPointF(-0.025, 0.0), QPointF(0.025, 0.0), QPointF(0.0, 0.025)])
        painter.drawPolygon(arrow.translated(bottom_arrow_point))
        t = QTransform()
        t.rotate(180.0 - 25.0)
        arrow_l = t.map(arrow)
        arrow_l = arrow_l.translated(curve_end_l)
        painter.drawPolygon(arrow_l)
        t = QTransform()
        t.rotate(180.0 + 25.0)
        arrow_r = t.map(arrow)
        arrow_r = arrow_r.translated(curve_end_r)
        painter.drawPolygon(arrow_r)


class TurboPumpSymbolIcon(BaseSymbolIcon):
    """
    A widget with a turbo pump symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._center_brush = QBrush(QColor("transparent"))

    @Property(QBrush)
    def centerBrush(self):
        return self._center_brush

    @centerBrush.setter
    def centerBrush(self, new_brush):
        if new_brush != self._center_brush:
            self._center_brush = new_brush
            self.update()

    def draw_icon(self, painter):
        # Outer circle
        painter.drawEllipse(QPointF(0.5, 0.5), 0.5, 0.5)

        brush = painter.brush()
        pen = painter.pen()

        painter.setBrush(self.centerBrush)

        # Inner concentric circles
        painter.drawEllipse(QPointF(0.5, 0.5), 0.2, 0.2)
        painter.drawEllipse(QPointF(0.5, 0.5), 0.1, 0.1)

        painter.setBrush(brush)
        painter.setPen(pen)

        # Inner straight lines
        painter.drawChord(QRectF(0.0, 0.0, 1.0, 1.0), 45 * 16, -120 * 16)
        painter.drawChord(QRectF(0.0, 0.0, 1.0, 1.0), 135 * 16, 120 * 16)


class GetterPumpSymbolIcon(BaseSymbolIcon):
    """
    A widget with a getter pump symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """

    def draw_icon(self, painter):
        painter.drawEllipse(QPointF(0.5, 0.5), 0.5, 0.5)
        painter.drawChord(QRectF(0.0, 0.0, 1.0, 1.0), 90 * 16, -100 * 16)
        painter.drawChord(QRectF(0.0, 0.0, 1.0, 1.0), 135 * 16, 100 * 16)

        # Draw the arrow end-caps
        painter.setBrush(QBrush(QColor(0, 0, 0)))

        top_arrow_point = QPointF(0.35, 0.15)
        arrow = QPolygonF(
            [QPointF(-0.08, 0.0),
             QPointF(-0.005, 0.0),
             QPointF(-0.005, 0.15),
             QPointF(0.005, 0.15),
             QPointF(0.005, 0.0),
             QPointF(0.08, 0.0),
             QPointF(0.00, -0.08)]
        )

        t = QTransform()
        t.rotate(-25)
        top_arrow_r = t.map(arrow)
        arrow_l = top_arrow_r.translated(top_arrow_point)
        painter.drawPolygon(arrow_l)

        bottom_left_arrow_point = QPointF(0.35, 0.89)
        t = QTransform()
        t.rotate(180.0 + 25.0)
        arrow_r = t.map(arrow)
        arrow_r = arrow_r.translated(bottom_left_arrow_point)
        painter.drawPolygon(arrow_r)

        bottom_right_arrow_point = QPointF(0.85, 0.65)
        t = QTransform()
        t.rotate(180.0 - 65.0)
        arrow_r = t.map(arrow)
        arrow_r = arrow_r.translated(bottom_right_arrow_point)
        painter.drawPolygon(arrow_r)



import os

from qtpy.QtWidgets import QApplication, QWidget

import pcdswidgets.icons as icons


def screenshot(widget, filename):
    """
    Takes a screenshot of the widget window, saves png image to file
    """
    s = str(filename) + ".png"
    print("Filename: ", s)
    p = QWidget.grab(widget)
    p.save(s, 'png')


app = QApplication([])

for ic_name in icons.__all__:
    ic = getattr(icons, ic_name)
    tp = ic()
    tp.setFixedSize(64, 64)
    tp.show()
    path = os.path.join(os.path.dirname(__file__), '..', 'docs', 'source',
                        '_static', "icons",
                        tp.__class__.__name__)

    screenshot(tp, path)



"""
Show a fully functional widget. Useful for development.

Invoke as e.g.
"python -m pcdswidgets.vacuum.demo PneumaticValveDA CRIX:VGC:11"
"""
import sys

import pydm
from pydm.utilities import setup_renderer

from ..gauges import *  # noqa
from ..others import *  # noqa
from ..pumps import *  # noqa
from ..valves import *  # noqa

cls = sys.argv[1]
prefix = sys.argv[2]
setup_renderer()

app = pydm.PyDMApplication(
    ui_file=None,
    hide_nav_bar=True,
    hide_menu_bar=True,
    hide_status_bar=True,
)

widget = globals()[cls]()
widget.channelsPrefix = 'ca://' + sys.argv[2]
app.main_window.set_display_widget(widget)
app.exec()



from unittest.mock import Mock

import pytest
from qtpy.QtCore import Qt
from qtpy.QtGui import QBrush, QColor

import pcdswidgets.icons
from pcdswidgets.icons.base import BaseSymbolIcon

icons = [getattr(pcdswidgets.icons, icon)
         for icon in pcdswidgets.icons.__all__]


@pytest.mark.parametrize('icon_class', icons, ids=pcdswidgets.icons.__all__)
def test_icon_smoke(qtbot, icon_class):
    icon = icon_class()
    qtbot.addWidget(icon)
    with qtbot.waitExposed(icon):
        icon.show()
    icon.repaint()


@pytest.fixture(scope='function')
def icon(qtbot):
    icon = BaseSymbolIcon()
    qtbot.addWidget(icon)
    return icon


@pytest.mark.parametrize('prop,value',
                         [('brush', QBrush(QColor(0, 0, 0), Qt.SolidPattern)),
                          ('penStyle', Qt.DotLine),
                          ('penColor', QColor(0, 0, 0)),
                          ('penWidth', 2.0)],
                         ids=('brush', 'penStyle', 'penColor', 'penWidth'))
def test_icon_properties(icon, prop, value):
    setattr(icon, prop, value)
    assert getattr(icon, prop) == value


def test_icon_clicks(qtbot, icon):
    mock = Mock()
    icon.clicked.connect(mock)
    qtbot.mouseClick(icon, Qt.LeftButton)
    assert mock.called



from pydm.utilities import is_qt_designer, remove_protocol
from qtpy.QtCore import Property, QEvent, QSize, Qt, Signal
from qtpy.QtGui import QBrush, QColor, QPainter, QPen
from qtpy.QtWidgets import (QApplication, QStyle, QStyleOption, QToolTip,
                            QWidget)

from ..utils import find_ancestor_for_widget


class BaseSymbolIcon(QWidget):
    """
    Base class to be used for all the Symbol Icon widgets.
    This class holds most of the properties to be exposed and takes care of 90%
    of the drawing code needed.

    Parameters
    ----------
    parent : QWidget
        The parent widget for this widget.
    """
    clicked = Signal()

    def __init__(self, parent=None):
        self._brush = QBrush(QColor(0, 255, 0), Qt.SolidPattern)
        self._original_brush = None
        self._rotation = 0
        self._pen_style = Qt.SolidLine
        self._pen = QPen(self._pen_style)
        self._pen.setCosmetic(True)
        self._pen_width = 1.0
        self._pen_color = QColor(0, 0, 0)
        self._pen.setWidthF(self._pen_width)
        self._pen.setColor(self._pen_color)
        self._original_pen_style = self._pen_style
        self._original_pen_color = self._pen_color
        super().__init__(parent=parent)
        self.setObjectName("icon")
        if not is_qt_designer():
            # We should  install the Event Filter only if we are running
            # and not at the Designer
            self.installEventFilter(self)

    def eventFilter(self, obj, event):
        """
        EventFilter to redirect "middle click" to :meth:`.show_address_tooltip`
        """
        # Override the eventFilter to capture all middle mouse button events,
        # and show a tooltip if needed.
        if event.type() == QEvent.MouseButtonPress:
            if event.button() == Qt.MiddleButton:
                self.show_state_channel(event)
                return True
        return False

    def show_state_channel(self, event):
        """
        Show the State Channel Tooltip and copy address to clipboard

        This is intended to replicate the behavior of the "middle click" from
        EDM. If the parent is not PCDSSymbolBase and does not have a valid
        State Channel nothing will be displayed.
        """
        from ..vacuum.base import PCDSSymbolBase

        p = find_ancestor_for_widget(self, PCDSSymbolBase)
        if not p:
            return

        state_suffix = getattr(p, '_state_suffix', None)
        if not state_suffix:
            return

        addr = f"{p.channelsPrefix}{state_suffix}"
        QToolTip.showText(event.globalPos(), addr)
        # If the address has a protocol, strip it out before putting it on the
        # clipboard.
        copy_text = remove_protocol(addr)

        clipboard = QApplication.clipboard()
        clipboard.setText(copy_text)
        event = QEvent(QEvent.Clipboard)
        QApplication.instance().sendEvent(clipboard, event)

    def minimumSizeHint(self):
        return QSize(32, 32)

    def paintEvent(self, event):
        """
        Paint events are sent to widgets that need to update themselves,
        for instance when part of a widget is exposed because a covering
        widget was moved.

        This method handles the painting with parameters from the stylesheet,
        configures the brush, pen and calls ```draw_icon``` so the specifics
        can be performed for each of the drawing classes.

        Parameters
        ----------
        event : QPaintEvent
        """
        opt = QStyleOption()
        opt.initFrom(self)
        painter = QPainter(self)
        painter.setClipping(True)
        self.style().drawPrimitive(QStyle.PE_Widget, opt, painter, self)
        painter.setRenderHint(QPainter.Antialiasing)
        w = self.width()
        h = self.height()
        painter.translate(w / 2.0, h / 2.0)
        painter.rotate(self._rotation)
        painter.translate(-w / 2.0, -h / 2.0)
        painter.translate(self._pen_width / 2.0, self._pen_width / 2.0)
        painter.scale(w - self._pen_width, h - self._pen_width)
        painter.translate(0, 0)
        painter.setBrush(self._brush)
        painter.setPen(self._pen)
        self.draw_icon(painter)

        QWidget.paintEvent(self, event)

    def draw_icon(self, painter):
        """
        Method responsible for the drawing of the icon part of the paintEvent.
        This method must be implemented by the symbol icon widgets to include
        their specific drawings.

        Parameters
        ----------
        painter : QPainter
        """
        raise NotImplementedError("draw_icon must be implemented by subclass")

    @Property(QBrush)
    def brush(self):
        """
        PyQT Property for the brush object to be used when coloring the
        drawing

        Returns
        -------
        QBrush
        """
        return self._brush

    @brush.setter
    def brush(self, new_brush):
        """
        PyQT Property for the brush object to be used when coloring the
        drawing

        Parameters
        ----------
        new_brush : QBrush
        """
        if new_brush != self._brush:
            self._brush = new_brush
            self.update()

    @Property(Qt.PenStyle)
    def penStyle(self):
        """
        PyQT Property for the pen style to be used when drawing the border

        Returns
        -------
        int
            Index at Qt.PenStyle enum
        """
        return self._pen_style

    @penStyle.setter
    def penStyle(self, new_style):
        """
        PyQT Property for the pen style to be used when drawing the border

        Parameters
        ----------
        new_style : int
            Index at Qt.PenStyle enum
        """
        if new_style != self._pen_style:
            self._pen_style = new_style
            self._pen.setStyle(new_style)
            self.update()

    @Property(QColor)
    def penColor(self):
        """
        PyQT Property for the pen color to be used when drawing the border

        Returns
        -------
        QColor
        """
        return self._pen_color

    @penColor.setter
    def penColor(self, new_color):
        """
        PyQT Property for the pen color to be used when drawing the border

        Parameters
        ----------
        new_color : QColor
        """
        if new_color != self._pen_color:
            self._pen_color = new_color
            self._pen.setColor(new_color)
            self.update()

    @Property(float)
    def penWidth(self):
        """
        PyQT Property for the pen width to be used when drawing the border

        Returns
        -------
        float
        """
        return self._pen_width

    @penWidth.setter
    def penWidth(self, new_width):
        """
        PyQT Property for the pen width to be used when drawing the border

        Parameters
        ----------
        new_width : float
        """
        if new_width < 0:
            return
        if new_width != self._pen_width:
            self._pen_width = int(new_width)
            self._pen.setWidth(self._pen_width)
            self.update()

    @Property(float)
    def rotation(self):
        """
        PyQt Property for the rotation.
        This rotates the icon coordinate system clockwise.

        Returns
        -------
        angle : float
            The angle in degrees
        """
        return self._rotation

    @rotation.setter
    def rotation(self, angle):
        """
        PyQt Property for the rotation.
        This rotates the icon coordinate system clockwise.

        Parameters
        ----------
        angle : float
            The angle in degrees
        """
        self._rotation = angle
        self.update()

    def mousePressEvent(self, evt):
        """Clicking the icon fires the ``clicked`` signal"""
        # Default Qt reaction to the icon press
        super().mousePressEvent(evt)
        if evt.button() == Qt.LeftButton:
            self.clicked.emit()



import logging

from qtpy.QtWidgets import QWidget

logger = logging.getLogger(__name__)


def refresh_style(widget):
    """
    Method that traverse the widget tree starting at `widget` and refresh the
    style for this widget and its childs.

    Parameters
    ----------
    widget : QWidget
    """
    widgets = [widget]
    widgets.extend(widget.findChildren(QWidget))
    for child_widget in widgets:
        child_widget.style().unpolish(child_widget)
        child_widget.style().polish(child_widget)
        child_widget.update()
        if child_widget != widget:
            refresh_style(child_widget)


def find_ancestor_for_widget(widget, klass):
    w = widget
    while w.parent() is not None:
        w = w.parent()
        if isinstance(w, klass):
            return w
    return None



import math

from qtpy.QtCore import Property, QLineF, QPointF, QRectF, Qt
from qtpy.QtGui import (QBrush, QColor, QPainterPath, QPen, QPolygonF,
                        QTransform)

from .base import BaseSymbolIcon


class PneumaticValveSymbolIcon(BaseSymbolIcon):
    """
    A widget with a pneumatic valve symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._interlock_brush = QBrush(QColor(0, 255, 0), Qt.SolidPattern)

    @Property(QBrush)
    def interlockBrush(self):
        return self._interlock_brush

    @interlockBrush.setter
    def interlockBrush(self, new_brush):
        if new_brush != self._interlock_brush:
            self._interlock_brush = new_brush
            self.update()

    def draw_icon(self, painter):
        path = QPainterPath(QPointF(0, 0.3))
        path.lineTo(0, 0.9)
        path.lineTo(1, 0.3)
        path.lineTo(1, 0.9)
        path.closeSubpath()
        painter.drawPath(path)
        painter.drawLine(QPointF(0.5, 0.6), QPointF(0.5, 0.3))
        painter.setBrush(self._interlock_brush)
        painter.drawRect(QRectF(0.2, 0, 0.6, 0.3))


class FastShutterSymbolIcon(BaseSymbolIcon):
    """
    A widget with a fast shutter symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._arrow_brush = QBrush(QColor("transparent"), Qt.SolidPattern)

    @Property(QBrush)
    def arrowBrush(self):
        return self._arrow_brush

    @arrowBrush.setter
    def arrowBrush(self, new_brush):
        if new_brush != self._arrow_brush:
            self._arrow_brush = new_brush
            self.update()

    def draw_icon(self, painter):
        path = QPainterPath(QPointF(0, 0.3))
        path.lineTo(0, 0.9)
        path.lineTo(1, 0.3)
        path.lineTo(1, 0.9)
        path.closeSubpath()
        painter.drawPath(path)

        prev_brush = painter.brush()
        prev_pen = painter.pen()

        painter.setPen(Qt.NoPen)
        painter.setBrush(self._arrow_brush)
        arrow = QPolygonF(
            [QPointF(0.2, 0),
             QPointF(0.2, 0.20),
             QPointF(0.5, 0.40),
             QPointF(0.8, 0.20),
             QPointF(0.8, 0)
             ]
        )
        painter.drawPolygon(arrow)

        painter.setPen(prev_pen)
        painter.setBrush(prev_brush)
        painter.drawLine(QPointF(0.2, 0), QPointF(0.5, 0.20))
        painter.drawLine(QPointF(0.2, 0.20), QPointF(0.5, 0.40))
        painter.drawLine(QPointF(0.5, 0.20), QPointF(0.8, 0))
        painter.drawLine(QPointF(0.5, 0.40), QPointF(0.8, 0.20))

        painter.drawLine(QPointF(0.5, 0.6), QPointF(0.5, 0.0))


class RightAngleManualValveSymbolIcon(BaseSymbolIcon):
    """
    A widget with a right angle manual valve symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def draw_icon(self, painter):
        path = QPainterPath(QPointF(0, 0))
        path.lineTo(1, 1)
        path.lineTo(0.005, 1)
        path.lineTo(0.5, 0.5)
        path.lineTo(0, 0.9)
        path.closeSubpath()
        painter.drawPath(path)
        painter.drawEllipse(QPointF(0.5, 0.5), 0.05, 0.05)


class ApertureValveSymbolIcon(BaseSymbolIcon):
    """
    A widget with an aperture valve symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._interlock_brush = QBrush(QColor(0, 255, 0), Qt.SolidPattern)

    @Property(QBrush)
    def interlockBrush(self):
        return self._interlock_brush

    @interlockBrush.setter
    def interlockBrush(self, new_brush):
        if new_brush != self._interlock_brush:
            self._interlock_brush = new_brush
            self.update()

    def draw_icon(self, painter):
        path = QPainterPath(QPointF(0, 0.3))
        path.lineTo(0, 0.9)
        path.lineTo(1, 0.3)
        path.lineTo(1, 0.9)
        path.closeSubpath()
        painter.drawPath(path)
        painter.drawEllipse(QPointF(0.5, 0.6), 0.1, 0.1)
        painter.drawLine(QPointF(0.5, 0.5), QPointF(0.5, 0.3))
        painter.setBrush(self._interlock_brush)
        painter.drawRect(QRectF(0.2, 0, 0.6, 0.3))


class NeedleValveSymbolIcon(BaseSymbolIcon):
    """
    A widget with a needle valve symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._interlock_brush = QBrush(QColor(0, 255, 0), Qt.SolidPattern)

    @Property(QBrush)
    def interlockBrush(self):
        return self._interlock_brush

    @interlockBrush.setter
    def interlockBrush(self, new_brush):
        if new_brush != self._interlock_brush:
            self._interlock_brush = new_brush
            self.update()

    def draw_icon(self, painter):
        path = QPainterPath(QPointF(0, 0.3))
        path.lineTo(0, 0.9)
        path.lineTo(1, 0.3)
        path.lineTo(1, 0.9)
        path.closeSubpath()
        painter.drawPath(path)
        painter.drawLine(QPointF(0.5, 0.6), QPointF(0.5, 0.15))

        # Draw the arrow end-caps
        painter.setBrush(QBrush(QColor(0, 0, 0)))

        top_arrow_point = QPointF(0.65, 0.36)
        arrow = QPolygonF(
            [QPointF(-0.09, 0.0),
             QPointF(-0.005, 0.0),
             QPointF(-0.005, 0.8),
             QPointF(0.005, 0.8),
             QPointF(0.005, 0.0),
             QPointF(0.09, 0.0),
             QPointF(0.00, -0.25)]
        )

        t = QTransform()
        t.rotate(35)
        top_arrow_r = t.map(arrow)
        arrow_l = top_arrow_r.translated(top_arrow_point)
        painter.drawPolygon(arrow_l)

        painter.setBrush(self._interlock_brush)
        painter.drawRect(QRectF(0.3, 0, 0.4, 0.15))


class ProportionalValveSymbolIcon(BaseSymbolIcon):
    """
    A widget with a proportional valve symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._interlock_brush = QBrush(QColor(0, 255, 0), Qt.SolidPattern)

    @Property(QBrush)
    def interlockBrush(self):
        return self._interlock_brush

    @interlockBrush.setter
    def interlockBrush(self, new_brush):
        if new_brush != self._interlock_brush:
            self._interlock_brush = new_brush
            self.update()

    def draw_icon(self, painter):
        path = QPainterPath(QPointF(0, 0.3))
        path.lineTo(0, 0.9)
        path.lineTo(1, 0.3)
        path.lineTo(1, 0.9)
        path.closeSubpath()
        painter.drawPath(path)
        painter.drawLine(QPointF(0.5, 0.6), QPointF(0.5, 0.15))

        painter.setBrush(self._interlock_brush)
        painter.drawRect(QRectF(0.35, 0, 0.3, 0.3))

        # Draw the arrow end-caps
        painter.setBrush(QBrush(QColor(0, 0, 0)))

        top_arrow_point = QPointF(0.65, 0.42)
        arrow = QPolygonF(
            [QPointF(-0.07, 0.0),
             QPointF(-0.005, 0.0),
             QPointF(-0.005, 0.8),
             QPointF(0.005, 0.8),
             QPointF(0.005, 0.0),
             QPointF(0.07, 0.0),
             QPointF(0.00, -0.25)]
        )

        t = QTransform()
        t.rotate(40)
        top_arrow_r = t.map(arrow)
        arrow_l = top_arrow_r.translated(top_arrow_point)
        painter.drawPolygon(arrow_l)

        t_x = 0.4
        t_y = 0.05
        painter.drawLines([QLineF(0.0+t_x, 0.0+t_y, 0.0+t_x, 0.2+t_y),
                           QLineF(0.0+t_x, 0.0+t_y, 0.1+t_x, 0.2+t_y),
                           QLineF(0.1+t_x, 0.2+t_y, 0.2+t_x, 0.0+t_y),
                           QLineF(0.2+t_x, 0.0+t_y, 0.2+t_x, 0.2+t_y)])


class ControlValveSymbolIcon(PneumaticValveSymbolIcon):
    """Icon for a Control Valve with readback"""
    def draw_icon(self, painter):
        pen = painter.pen()
        pen.setWidthF(pen.width()*2)
        pen.setCapStyle(Qt.FlatCap)
        painter.setPen(pen)
        # Circle parameters
        radius = 0.3
        center = (0.5, 1 - radius)
        # Draw circle
        painter.drawEllipse(QPointF(*center),
                            radius, radius)
        # X pattern
        quad = math.cos(math.radians(45)) * radius
        painter.drawLine(QLineF(center[0] + quad,
                                center[1] + quad,
                                center[0] - quad,
                                center[1] - quad))
        painter.drawLine(QLineF(center[0] + quad,
                                center[1] - quad,
                                center[0] - quad,
                                center[1] + quad))
        # Interlock Icon
        square_dims = (0.4, 0.2)
        painter.drawLine(QPointF(center[0], center[1] - radius),
                         QPointF(center[0], square_dims[1]))
        painter.setBrush(self._interlock_brush)
        painter.drawRect(QRectF((1 - square_dims[0])/2., 0, *square_dims))


class ControlOnlyValveSymbolIcon(BaseSymbolIcon):
    """Icon for a Control Valve with no readback"""
    def draw_icon(self, painter):
        path = QPainterPath(QPointF(0, 0.3))
        path.lineTo(0, 0.9)
        path.lineTo(1, 0.3)
        path.lineTo(1, 0.9)
        path.closeSubpath()
        painter.drawPath(path)


class PneumaticValveNOSymbolIcon(BaseSymbolIcon):
    """
    A widget with a normally open pneumatic valve symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._interlock_brush = QBrush(QColor(0, 255, 0), Qt.SolidPattern)

    @Property(QBrush)
    def interlockBrush(self):
        return self._interlock_brush

    @interlockBrush.setter
    def interlockBrush(self, new_brush):
        if new_brush != self._interlock_brush:
            self._interlock_brush = new_brush
            self.update()

    def draw_icon(self, painter):
        path = QPainterPath(QPointF(0, 0.3))
        path.lineTo(0, 0.9)
        path.lineTo(1, 0.3)
        path.lineTo(1, 0.9)
        path.closeSubpath()
        painter.drawPath(path)
        painter.drawLine(QPointF(0.5, 0.6), QPointF(0.5, 0.3))
        painter.setBrush(self._interlock_brush)
        painter.drawRect(QRectF(0.2, 0, 0.6, 0.3))
        # Draw the N
        n_path = QPainterPath(QPointF(0.25, 0.25))
        n_path.lineTo(0.25, 0.05)
        n_path.lineTo(0.45, 0.25)
        n_path.lineTo(0.45, 0.05)
        painter.drawPath(n_path)

        # Draw the O
        painter.drawEllipse(QPointF(0.65, 0.15), 0.1, 0.1)


class PneumaticValveDASymbolIcon(BaseSymbolIcon):
    """
    A widget with a dual-acting pneumatic valve symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)
        self._interlock_brush = QBrush(QColor(0, 255, 0), Qt.SolidPattern)

    @Property(QBrush)
    def interlockBrush(self):
        return self._interlock_brush

    @interlockBrush.setter
    def interlockBrush(self, new_brush):
        if new_brush != self._interlock_brush:
            self._interlock_brush = new_brush
            self.update()

    def draw_icon(self, painter):
        path = QPainterPath(QPointF(0, 0.3))
        path.lineTo(0, 0.9)
        path.lineTo(1, 0.3)
        path.lineTo(1, 0.9)
        path.closeSubpath()
        painter.drawPath(path)
        painter.drawLine(QPointF(0.5, 0.6), QPointF(0.5, 0.3))
        painter.setBrush(self._interlock_brush)
        painter.drawRect(QRectF(0.2, 0, 0.6, 0.3))

        # Set fill color to black, line width to minimum for arrows
        black_brush = QBrush(QColor(0, 0, 0))
        painter.setBrush(black_brush)
        painter.setPen(QPen(black_brush, 0))

        # Draw an arrow around 0, 0 pointing right
        # This polygon starts from the tip and works its way around clockwise
        tip_length = 0.2
        tip_width = 0.1
        length = 0.2
        width = 0.02
        rightward_arrow = QPolygonF(
            [QPointF(tip_length, 0.0),
             QPointF(0.0, -tip_width),
             QPointF(0.0, -width),
             QPointF(-length, -width),
             QPointF(-length, width),
             QPointF(0.0, width),
             QPointF(0.0, tip_width)]
        )
        # Second arrow looks left
        point_left = QTransform()
        point_left.rotate(180)
        leftward_arrow = point_left.map(rightward_arrow)

        # Establish start positions for the arrows
        # This is where the origin goes (0, 0)
        # This is where the triangle meets the line
        right_start = QPointF(0.59, 0.15)
        left_start = QPointF(0.41, 0.15)

        # Assign arrows to positions
        right_translated = rightward_arrow.translated(right_start)
        left_translated = leftward_arrow.translated(left_start)

        # Call draw
        painter.drawPolygon(right_translated)
        painter.drawPolygon(left_translated)



import pytest

import pcdswidgets.vacuum

symbols = [getattr(pcdswidgets.vacuum, symbol)
           for symbol in pcdswidgets.vacuum.__all__]


@pytest.mark.parametrize('symbol', symbols, ids=pcdswidgets.vacuum.__all__)
def test_vacuum_widgets(qtbot, symbol):
    widget = symbol()
    qtbot.addWidget(widget)
    widget.create_channels()
    widget.destroy_channels()



from qtpy.QtCore import QPointF, QRectF
from qtpy.QtGui import QPainterPath

from .base import BaseSymbolIcon


class RoughGaugeSymbolIcon(BaseSymbolIcon):
    """
    A widget with a pirani gauge symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """

    path = QPainterPath(QPointF(0.5, 0))
    path.lineTo(1, 1)
    path.lineTo(0, 1)
    path.closeSubpath()

    def draw_icon(self, painter):
        painter.drawPath(self.path)


class CathodeGaugeSymbolIcon(BaseSymbolIcon):
    """
    A widget with a cathode gauge symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """

    def draw_icon(self, painter):
        painter.drawEllipse(QPointF(0.5, 0.5), 0.5, 0.5)


class HotCathodeGaugeSymbolIcon(CathodeGaugeSymbolIcon):
    """
    A widget with a hot cathode gauge symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def draw_icon(self, painter):
        super().draw_icon(painter)
        painter.drawLine(QPointF(0.3, 0.1), QPointF(0.3, 0.9))
        painter.drawLine(QPointF(0.3, 0.5), QPointF(0.7, 0.5))
        painter.drawLine(QPointF(0.7, 0.1), QPointF(0.7, 0.9))


class ColdCathodeGaugeSymbolIcon(CathodeGaugeSymbolIcon):
    """
    A widget with a cold cathode gauge symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    def draw_icon(self, painter):
        super().draw_icon(painter)
        painter.drawArc(QRectF(0.25, 0.25, 0.5, 0.5), 45*16, 270*16)



from .gauges import (CathodeGaugeSymbolIcon, ColdCathodeGaugeSymbolIcon,
                     HotCathodeGaugeSymbolIcon, RoughGaugeSymbolIcon)
from .others import RGASymbolIcon
from .pumps import (GetterPumpSymbolIcon, IonPumpSymbolIcon,
                    ScrollPumpSymbolIcon, TurboPumpSymbolIcon)
from .valves import (ApertureValveSymbolIcon, ControlOnlyValveSymbolIcon,
                     ControlValveSymbolIcon, FastShutterSymbolIcon,
                     NeedleValveSymbolIcon, PneumaticValveDASymbolIcon,
                     PneumaticValveNOSymbolIcon, PneumaticValveSymbolIcon,
                     ProportionalValveSymbolIcon,
                     RightAngleManualValveSymbolIcon)

__all__ = ['RoughGaugeSymbolIcon', 'CathodeGaugeSymbolIcon',
           'HotCathodeGaugeSymbolIcon', 'ColdCathodeGaugeSymbolIcon',
           'IonPumpSymbolIcon', 'TurboPumpSymbolIcon', 'ScrollPumpSymbolIcon',
           'GetterPumpSymbolIcon',
           'PneumaticValveSymbolIcon', 'FastShutterSymbolIcon',
           'ApertureValveSymbolIcon', 'RightAngleManualValveSymbolIcon',
           'NeedleValveSymbolIcon', 'ProportionalValveSymbolIcon',
           'RGASymbolIcon', 'ControlValveSymbolIcon',
           'ControlOnlyValveSymbolIcon', 'PneumaticValveNOSymbolIcon',
           'PneumaticValveDASymbolIcon']



import pytest
from qtpy.QtWidgets import QHBoxLayout, QVBoxLayout

from pcdswidgets.icons import RGASymbolIcon
from pcdswidgets.vacuum.base import ContentLocation, PCDSSymbolBase


class BaseSymbol(PCDSSymbolBase):
    """Test Symbol for base class tests"""
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.icon = RGASymbolIcon(parent=self)


@pytest.fixture(scope='function')
def symbol(qtbot):
    symbol = BaseSymbol()
    qtbot.addWidget(symbol)
    return symbol


def test_show_icon(symbol):
    symbol.showIcon = False
    assert not symbol.icon.isVisible()


def test_symbol_paintEvent_smoke(symbol):
    symbol.show()


def test_no_controls_content(symbol):
    symbol.controlsLocation = ContentLocation.Hidden
    widget_layout = symbol.interlock.layout().itemAt(0).layout()
    widget = widget_layout.itemAt(0).widget()
    assert widget == symbol.icon


@pytest.mark.parametrize('location,layout,position',
                         [(ContentLocation.Top, QVBoxLayout, 0),
                          (ContentLocation.Bottom, QVBoxLayout, 1),
                          (ContentLocation.Left, QHBoxLayout, 0),
                          (ContentLocation.Right, QHBoxLayout, 1)],
                         ids=['Top', 'Bottom', 'Left', 'Right'])
def test_controls_content_location(symbol, location, layout, position):
    symbol.controlsLocation = location
    assert isinstance(symbol.interlock.layout(), layout)
    widget_layout = symbol.interlock.layout().itemAt(position).layout()
    widget = widget_layout.itemAt(0).widget()
    assert widget == symbol.controls_frame


def test_icon_fixed_size(symbol):
    size = 30
    symbol.iconSize = size
    assert symbol.icon.width() == size
    assert symbol.icon.height() == size


@pytest.mark.parametrize('rotate', (False, True), ids=('Standard', 'Rotated'))
def test_icon_rotation(symbol, rotate):
    symbol.rotateIcon = rotate
    assert symbol.icon.rotation == 90 * int(rotate)



"""
Show an icon. Useful for development.

Invoke as e.g. "python -m pcdswidgets.icons.demo ControlValve"
"""
import sys

from qtpy.QtWidgets import QApplication

from .. import *  # noqa

cls = sys.argv[1] + 'SymbolIcon'
app = QApplication([])
icon = globals()[cls]()
icon.show()
app.exec()



from pydm.widgets.display_format import DisplayFormat
from qtpy.QtCore import QSize

from ..icons.gauges import (ColdCathodeGaugeSymbolIcon,
                            HotCathodeGaugeSymbolIcon, RoughGaugeSymbolIcon)
from .base import PCDSSymbolBase
from .mixins import (ButtonLabelControl, InterlockMixin, LabelControl,
                     StateMixin)


class RoughGauge(StateMixin, LabelControl, PCDSSymbolBase):
    """
    A Symbol Widget representing a Rough Gauge with the proper icon and
    controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+
    |pressure   |PyDMLabel     |The pressure reading label.            |
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------------+
    |Property   |Values                                                       |
    +===========+=============================================================+
    |state      |`On` or `Off`                                                |
    +-----------+-------------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        RoughGauge[state="Off"] {
            qproperty-brush: red;
            color: gray;
        }
        RoughGauge[state="On"] {
            qproperty-brush: green;
            color: black;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Gauges",
        "is_container": False,
    }
    _state_suffix = ":STATE_RBV"
    _readback_suffix = ":PRESS_RBV"

    NAME = "Rough Gauge"
    EXPERT_OPHYD_CLASS = "pcdsdevices.gauge.GaugePLC"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            state_suffix=self._state_suffix,
            readback_suffix=self._readback_suffix,
            readback_name='pressure',
            **kwargs)
        self.icon = RoughGaugeSymbolIcon(parent=self)
        self.readback_label.displayFormat = DisplayFormat.Exponential

    def sizeHint(self):
        return QSize(70, 60)


class HotCathodeGauge(ButtonLabelControl, InterlockMixin, StateMixin, PCDSSymbolBase):
    """
    A Symbol Widget representing a Hot Cathode Gauge with the proper icon
    and controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+
    |pressure   |PyDMLabel     |The pressure reading label.            |
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------------+
    |Property   |Values                                                       |
    +===========+=============================================================+
    |interlocked|`true` or `false`                                            |
    +-----------+-------------------------------------------------------------+
    |state      |`On`, `Off`, `Starting` or `Error`                           |
    +-----------+-------------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        HotCathodeGauge[interlocked="true"] #interlock {
            border: 5px solid red;
        }
        HotCathodeGauge[interlocked="false"] #interlock {
            border: 0px;
        }
        HotCathodeGauge[state="Error"] #icon {
            qproperty-penColor: red;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Gauges",
        "is_container": False,
    }
    _interlock_suffix = ":ILK_OK_RBV"
    _state_suffix = ":STATE_RBV"
    _readback_suffix = ":PRESS_RBV"
    _command_suffix = ":HV_SW"

    NAME = "Hot Cathode Gauge"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            readback_suffix=self._readback_suffix,
            readback_name='pressure',
            **kwargs)
        self.icon = HotCathodeGaugeSymbolIcon(parent=self)
        self.readback_label.displayFormat = DisplayFormat.Exponential

    def sizeHint(self):
        return QSize(180, 80)


class ColdCathodeGauge(InterlockMixin, StateMixin, ButtonLabelControl, PCDSSymbolBase):
    """
    A Symbol Widget representing a Cold Cathode Gauge with the proper icon and
    controls.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the symbol

    Notes
    -----
    This widget allow for high customization through the Qt Stylesheets
    mechanism.
    As this widget is composed by internal widgets, their names can be used as
    selectors when writing your stylesheet to be used with this widget.
    Properties are also available to offer wider customization possibilities.

    **Internal Components**

    +-----------+--------------+---------------------------------------+
    |Widget Name|Type          |What is it?                            |
    +===========+==============+=======================================+
    |interlock  |QFrame        |The QFrame wrapping this whole widget. |
    +-----------+--------------+---------------------------------------+
    |controls   |QFrame        |The QFrame wrapping the controls panel.|
    +-----------+--------------+---------------------------------------+
    |icon       |BaseSymbolIcon|The widget containing the icon drawing.|
    +-----------+--------------+---------------------------------------+
    |pressure   |PyDMLabel     |The pressure reading label.            |
    +-----------+--------------+---------------------------------------+

    **Additional Properties**

    +-----------+-------------------------------------------------------------+
    |Property   |Values                                                       |
    +===========+=============================================================+
    |interlocked|`true` or `false`                                            |
    +-----------+-------------------------------------------------------------+
    |state      |`On`, `Off`, `Starting` or `Error`                           |
    +-----------+-------------------------------------------------------------+

    Examples
    --------

    .. code-block:: css

        ColdCathodeGauge[interlocked="true"] #interlock {
            border: 5px solid red;
        }
        ColdCathodeGauge[interlocked="false"] #interlock {
            border: 0px;
        }
        ColdCathodeGauge[state="Error"] #icon {
            qproperty-penColor: red;
            qproperty-penWidth: 2;
        }

    """

    _qt_designer_ = {
        "group": "PCDS Gauges",
        "is_container": False,
    }
    _interlock_suffix = ":ILK_OK_RBV"
    _state_suffix = ":STATE_RBV"
    _readback_suffix = ":PRESS_RBV"
    _command_suffix = ":HV_SW"

    NAME = "Cold Cathode Gauge"
    EXPERT_OPHYD_CLASS = "pcdsdevices.gauge.GCCPLC"

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            interlock_suffix=self._interlock_suffix,
            state_suffix=self._state_suffix,
            command_suffix=self._command_suffix,
            readback_suffix=self._readback_suffix,
            readback_name='pressure',
            **kwargs)
        self.icon = ColdCathodeGaugeSymbolIcon(parent=self)
        self.readback_label.displayFormat = DisplayFormat.Exponential

    def sizeHint(self):
        return QSize(180, 80)



import pytest
from qtpy.QtWidgets import QWidget

from pcdswidgets.vacuum.base import PCDSSymbolBase
from pcdswidgets.vacuum.mixins import (ErrorMixin, InterlockMixin,
                                       OpenCloseStateMixin, StateMixin)


class PCDSSymbolWithIcon(PCDSSymbolBase):
    """Base mixable class"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.icon = QWidget(parent=self)


class Interlock(InterlockMixin, PCDSSymbolWithIcon):
    """Simplest Interlock Widget"""
    pass


class Error(ErrorMixin, PCDSSymbolWithIcon):
    """Simplest Error Widget"""
    pass


class State(StateMixin, PCDSSymbolWithIcon):
    """Simplest State Widget"""
    pass


class OpenClose(OpenCloseStateMixin, PCDSSymbolWithIcon):
    """Simplest OpenCloseState Widget"""
    pass


@pytest.fixture(scope='function')
def interlock(qtbot):
    inter = Interlock(':ILK')
    qtbot.addWidget(inter)
    inter.create_channels()
    return inter


@pytest.fixture(scope='function')
def error(qtbot):
    error = Error(':ILK')
    qtbot.addWidget(error)
    error.create_channels()
    error.error_enum_changed(('Bad', 'Good'))
    return error


@pytest.fixture(scope='function')
def state(qtbot):
    state = State(':Status')
    qtbot.addWidget(state)
    state.create_channels()
    state.state_enum_changed(('Bad', 'Good'))
    return state


@pytest.fixture(scope='function')
def openclose(qtbot):
    openclose = OpenClose(':Open', 'Close')
    qtbot.addWidget(openclose)
    openclose.create_channels()
    return openclose


@pytest.mark.parametrize('interlock_bit',
                         (0, 1),
                         ids=('low', 'high'))
def test_interlock_value_changed(interlock, interlock_bit):
    interlock.interlock_value_changed(not interlock_bit)
    orig_tooltip = interlock.status_tooltip()
    interlock.interlock_value_changed(interlock_bit)
    interlocked = not bool(interlock_bit)
    assert interlock.interlocked == interlocked
    assert interlock.controls_frame.isEnabled() != interlocked
    assert interlock.status_tooltip() != orig_tooltip


def test_error_value_changed(error):
    orig_tooltip = error.status_tooltip()
    error.error_value_changed(1)
    error.error == 'Good'
    assert orig_tooltip != error.status_tooltip()


def test_state_value_changed(state):
    orig_tooltip = state.status_tooltip()
    state.state_value_changed(1)
    state.state == 'Good'
    assert orig_tooltip != state.status_tooltip()


@pytest.mark.parametrize('open_switch,closed_switch,state',
                         [(1, 1, 'INVALID'), (1, 0, 'Open'),
                          (0, 1, 'Close'), (0, 0, 'INVALID')],
                         ids=['Fault', 'Open', 'Closed', 'Invalid'])
def test_openclose_value_changed(openclose, open_switch, closed_switch, state):
    openclose.state_value_changed('OPEN', open_switch)
    openclose.state_value_changed('CLOSE', closed_switch)
    assert openclose.state == state



from __future__ import annotations

import dataclasses
import functools
import json
import logging
from typing import Any, Callable

from pydm.utilities import is_qt_designer
from pydm.widgets import PyDMEmbeddedDisplay
from pydm.widgets.channel import PyDMChannel
from qtpy import QtCore, QtGui, QtWidgets

logger = logging.getLogger(__name__)


class FilterSortWidgetTable(QtWidgets.QTableWidget):
    """
    Displays repeated widgets that are sortable and filterable.

    This will allow you to sort or filter based on macros and based on the
    values in each pydm widget.
    """
    _qt_designer_ = {
        "group": "PCDS Utilities",
        "is_container": False,
    }

    # Public instance variables
    template_widget: PyDMEmbeddedDisplay

    # Private instance variables
    _ui_filename: str | None
    _macros_filename: str | None
    _macros: list[dict[str, str]]
    _channel_headers: list[str]
    _macro_headers: list[str]
    _header_map: dict[str, int]
    _channels: list[PyDMChannel]
    _filters: dict[str, FilterInfo]
    _initial_sort_header: str
    _initial_sort_ascend: bool
    _hide_headers: list[str]
    _configurable: bool
    _watching_cells: bool

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._ui_filename = None
        self._macros_filename = None
        self.template_widget = PyDMEmbeddedDisplay(parent=self)
        self.template_widget.hide()
        self.template_widget.loadWhenShown = False
        self._macros = []
        self._channel_headers = []
        self._macro_headers = []
        self._header_map = {}
        self._channels = []
        self._filters = {}
        self._initial_sort_header = 'index'
        self._initial_sort_ascend = True
        self._hide_headers = []

        # Table settings
        self.setShowGrid(True)
        self.setSortingEnabled(False)
        self.setSelectionMode(self.NoSelection)
        self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().hide()

        self.configurable = False

        self._watching_cells = False

    def channels(self) -> list[PyDMChannel]:
        """
        Tell PyDM about our table channels so it knows to close them at exit.
        """
        return self._channels

    @QtCore.Property(str)
    def ui_filename(self) -> str:
        """
        Name of the ui file that is to be repeated to fill the table.

        This is currently required. When this is changed, we'll rebuild the
        table.
        """
        return self._ui_filename

    @ui_filename.setter
    def ui_filename(self, filename: str):
        self._ui_filename = filename
        self.reload_ui_file()
        self.reinit_table()

    def reload_ui_file(self) -> None:
        """
        Load the UI file and inspect it for PyDM channels.
        """
        try:
            self.template_widget.filename = self.ui_filename
        except Exception:
            logger.exception(
                "Reloading the UI file %s failed",
                self.ui_filename,
            )
            return
        # Let's find all the widgets with channels and save their names
        self._channel_headers = []
        for widget in self.template_widget.embedded_widget.children():
            try:
                ch = widget.channels()
            except Exception:
                # It is expected that some widgets do not have channels
                continue
            if ch:
                self._channel_headers.append(widget.objectName())

    @QtCore.Property(str)
    def macros_filename(self) -> str:
        """
        Json file defining PyDM macros. Optional.

        This follows the same format as used for the PyDM Template Repeater.
        If omitted, you should pass in macros using the set_macros method
        instead.
        """
        return self._macros_filename

    @macros_filename.setter
    def macros_filename(self, filename: str):
        self._macros_filename = filename
        self.reload_macros_file()

    def reload_macros_file(self) -> None:
        """
        Load the macros_filename and call set_macros.
        """
        if not self.macros_filename:
            return
        try:
            with open(self.macros_filename) as fd:
                macros = json.load(fd)
            self.set_macros(macros)
        except Exception:
            logger.exception('')
            return

    def set_macros(self, macros_list: list[dict[str, str]]) -> None:
        """
        Change the PyDM macros we use to load the table widgets.

        This causes the table to be rebuilt.

        Parameters
        ----------
        macros_list : list of dict
            A list where each element is a dictionary that defines the macros
            to pass in to one instance of the repeated widget. All dicts must
            have the same keys or this will not work properly.
        """
        self._macros = macros_list
        self._macro_headers = (
            list(self._macros[0].keys())
            if self._macros
            else []
        )
        self.reinit_table()

    def reinit_table(self) -> None:
        """
        Rebuild the table based on the ui_filename and the newest macros.
        """
        if self._watching_cells:
            self.cellChanged.disconnect(self.handle_item_changed)
            self._watching_cells = False
        for channel in self._channels:
            channel.disconnect()
        self._channels = []
        self.clear()
        self.clearContents()
        self.setRowCount(0)
        self._header_map = {}
        if not self._macros and self._channel_headers:
            return
        # Column 1 displays widget, 2 is index, the rest hold values
        ncols = 2 + len(self._channel_headers) + len(self._macro_headers)
        self.setColumnCount(ncols)
        for col in range(1, ncols):
            self.hideColumn(col)
        for macros in self._macros:
            self.add_row(macros)

        self._watching_cells = True
        self.cellChanged.connect(self.handle_item_changed)
        self.update_all_filters()

    def add_row(self, macros: dict[str, str]) -> None:
        """
        Adds a single row to the table.

        Each row will be created from the same UI file template.
        The macros used must have the same keys as all the previously
        added rows, or else the table will not work correctly.

        Parameters
        ----------
        macros : dict of str
            The macro substitutions for the UI file. These must be
            strings because we're effectively substituting them into
            the file's text.
        """
        widget = PyDMEmbeddedDisplay(parent=self)
        widget.macros = json.dumps(macros)
        widget.filename = self.ui_filename
        widget.loadWhenShown = False
        widget.disconnectWhenHidden = False
        self.add_context_menu_to_children(widget.embedded_widget)

        row_position = self.rowCount()
        self.insertRow(row_position)

        # Put the widget into the table
        self.setCellWidget(row_position, 0, widget)
        self._header_map['widget'] = 0
        self.setRowHeight(row_position, widget.height())

        # Put the index into the table
        item = ChannelTableWidgetItem(
            header='index',
            default=row_position,
        )
        self.setItem(row_position, 1, item)
        self._header_map['index'] = 1
        # Put the macros into the table
        index = 2
        for key, value in macros.items():
            item = ChannelTableWidgetItem(
                header=key,
                default=value,
            )
            self.setItem(row_position, index, item)
            self._header_map[key] = index
            index += 1
        # Set up the data columns and the channels
        for header in self._channel_headers:
            source = widget.findChild(QtCore.QObject, header)
            item = ChannelTableWidgetItem(
                header=header,
                channel=source.channel,
            )
            self.setItem(row_position, index, item)
            self._header_map[header] = index
            if item.pydm_channel is not None:
                self._channels.append(item.pydm_channel)
            index += 1

    def add_context_menu_to_children(self, widget: QtWidgets.QWidget) -> None:
        """
        Distribute the context menu to child widgets.

        This makes it so you can right click to configure the table from
        within any of the contained widgets.
        """
        for widget in widget.children():
            widget.contextMenuEvent = self.contextMenuEvent

    def contextMenuEvent(self, _event) -> None:
        """
        On right click, create and open a settings menu.
        """
        menu = QtWidgets.QMenu(parent=self)
        configure_action = menu.addAction('Configure')
        configure_action.setCheckable(True)
        configure_action.setChecked(self.configurable)
        configure_action.toggled.connect(self.request_configurable)
        active_sort_action = menu.addAction('Active Re-sort')
        active_sort_action.setCheckable(True)
        active_sort_action.setChecked(self.isSortingEnabled())
        active_sort_action.toggled.connect(self.setSortingEnabled)
        sort_menu = menu.addMenu('Sorting')
        for header_name in self._header_map.keys():
            if header_name == 'widget':
                continue
            if header_name in self.hide_headers_in_menu:
                continue
            inner_menu = sort_menu.addMenu(header_name.lower())
            asc = inner_menu.addAction('Ascending')
            asc.triggered.connect(
                functools.partial(
                    self.menu_sort,
                    header=header_name,
                    ascending=True,
                )
            )
            dec = inner_menu.addAction('Descending')
            dec.triggered.connect(
                functools.partial(
                    self.menu_sort,
                    header=header_name,
                    ascending=False,
                )
            )
        filter_menu = menu.addMenu('Filters')
        for filter_name, filter_info in self._filters.items():
            inner_action = filter_menu.addAction(filter_name)
            inner_action.setCheckable(True)
            inner_action.setChecked(filter_info.active)
            inner_action.toggled.connect(
                functools.partial(
                    self.activate_filter,
                    filter_name=filter_name,
                )
            )
        menu.exec_(QtGui.QCursor.pos())

    def get_row_values(self, row: int) -> dict[str, Any]:
        """
        Get the current values for a specific numbered row of the table.

        Parameters
        ----------
        row : int
            The row index to inspect. 0 is the current top row.

        Returns
        -------
        values : dict
            A mapping from str to value for each named widget in the template
            that has a PyDM channel. There is one additional special str, which
            is the 'connected' str, which is True if all channels are
            connected.
        """
        values = {'connected': True}
        for col in range(1, self.columnCount()):
            item = self.item(row, col)
            values[item.header] = item.get_value()
            if not item.connected:
                values['connected'] = False
        return values

    def add_filter(
        self,
        filter_name: str,
        filter_func: Callable[[dict[str, Any]], bool],
        active: bool = True
    ) -> None:
        """
        Add a new visibility filter to the table.

        Filters are functions with the following signature:
        ``filt(values: dict[str, Any]) -> bool``
        Where values is the output from get_row_values,
        and the boolean return value is True if the row should be displayed.
        If we have multiple filters, we need all of them to be True to display
        the row.

        Parameters
        ----------
        filter_name : str
            A name assigned to the filter to help us keep track of it.
        filter_func : func
            A callable with the correct signature.
        active : bool, optional.
            True if we want the filter to start as active. An inactive filter
            does not act on the table until the user requests it from the
            right-click context menu. Defaults to True.
        """
        # Filters take in a dict of values from header to value
        # Return True to show, False to hide
        self._filters[filter_name] = FilterInfo(
            filter_func=filter_func,
            active=active,
            name=filter_name,
        )
        self.update_all_filters()

    def remove_filter(self, filter_name: str) -> None:
        """
        Remove a specific named visibility filter from the table.

        This is a filter that was previously added using add_filter.

        Parameters
        ----------
        filter_name : str
            A name assigned to the filter to help us keep track of it.
        """
        del self._filters[filter_name]
        self.update_all_filters()

    def clear_filters(self) -> None:
        """
        Remove all visbility filters from the table.
        """
        self._filters = {}
        self.update_all_filters()

    def update_all_filters(self) -> None:
        """
        Apply all filters to all rows of the table.
        """
        for row in range(self.rowCount()):
            self.update_filter(row)

    def update_filter(self, row: int) -> None:
        """
        Apply all filters to one row of the table.

        Parameters
        ----------
        row : int
            The row index to inspect. 0 is the current top row.
        """
        if self._filters:
            values = self.get_row_values(row)
            show_row = []
            for filt_info in self._filters.values():
                if filt_info.active:
                    try:
                        should_show = filt_info.filter_func(values)
                    except Exception:
                        logger.debug(
                            'Error in filter function %s',
                            filt_info.name,
                            exc_info=True,
                        )
                        should_show = True
                    show_row.append(should_show)
                else:
                    # If inactive, record it as unfiltered/shown
                    show_row.append(True)
            if all(show_row):
                self.showRow(row)
            else:
                self.hideRow(row)
        else:
            self.showRow(row)

    def activate_filter(self, active: bool, filter_name: str) -> None:
        """
        Activate or deactivate a filter by name.

        Parameters
        ----------
        active : bool
            True if we want to activate a filter, False if we want to
            deactivate it.
        filter_name : str
            The name associated with the filter, chosen when we add the filter
            to the table.
        """
        self._filters[filter_name].active = active
        self.update_all_filters()

    def handle_item_changed(self, row: int, col: int) -> None:
        """
        Slot that is run when any element in the table changes.

        Currently, this updates the filters for the row that changed.
        """
        self.update_filter(row)

    @QtCore.Property(str)
    def initial_sort_header(self) -> str:
        """
        Column to sort on after initializing.

        Use this to initialize the sort order rather than using the sort_table
        function.
        """
        return self._initial_sort_header

    @initial_sort_header.setter
    def initial_sort_header(self, header: str):
        self._initial_sort_header = header
        if not is_qt_designer():
            # Do a sort after a short timer
            # HACK: this is because it doesn't work if done immediately
            # This is due to some combination of qt designer properties being
            # applied in some random order post-__init__ combined with it
            # taking a short bit of time for the items to be ready to sort.
            # Some items will never connect and never be ready to sort,
            # so for now we'll do a best-effort one-second wait.
            timer = QtCore.QTimer(parent=self)
            timer.singleShot(1000, self.initial_sort)

    @QtCore.Property(bool)
    def initial_sort_ascending(self) -> bool:
        """
        Whether to do the initial sort in ascending or descending order.
        """
        return self._initial_sort_ascend

    @initial_sort_ascending.setter
    def initial_sort_ascending(self, ascending: bool):
        self._initial_sort_ascend = ascending

    def initial_sort(self) -> None:
        """
        Called if the user specifies an initial_sort_header.
        """
        self.sort_table(self.initial_sort_header, self.initial_sort_ascending)

    @QtCore.Property('QStringList')
    def hide_headers_in_menu(self) -> list[str]:
        """
        A list of headers that we don't want to see in the sort menu.
        """
        return self._hide_headers

    @hide_headers_in_menu.setter
    def hide_headers_in_menu(self, headers: list[str]):
        self._hide_headers = headers

    def sort_table(self, header: str, ascending: bool) -> None:
        """
        Rearrange the ordering of the table based on any of the value fields.

        Parameters
        ----------
        header : str
            The name of any of the value fields to use to sort on. Valid
            headers are 'index', which is the original sort order, strings that
            match the macro keys, and strings that match widget names in the
            template.
        ascending : bool
            If True, we'll sort in ascending order. If False, we'll sort in
            descending order.
        """
        self.reset_manual_sort()
        if ascending:
            order = QtCore.Qt.AscendingOrder
        else:
            order = QtCore.Qt.DescendingOrder
        col = self._header_map[header]
        self.sortByColumn(col, order)

    def menu_sort(self, checked: bool, header: str, ascending: bool):
        """
        sort_table wrapped to recieve the checked bool from a signal.

        Ignore the checked boolean.
        """
        self.sort_table(header, ascending)

    def reset_manual_sort(self) -> None:
        """
        Rearrange the table to undo all manual drag/drop sorting.
        """
        header = self.verticalHeader()
        for row in range(self.rowCount()):
            header.moveSection(header.visualIndex(row), row)

    @QtCore.Property(bool, designable=False)
    def configurable(self) -> bool:
        """
        Whether or not the table can be manipulated from the UI.

        If True, the table rows can be dragged/dropped/rearranged.
        If False, the table rows can no longer be selected.

        This begins as False if unset and can be changed in the context menu.
        """
        return self._configurable

    @configurable.setter
    def configurable(self, conf: bool):
        self._configurable = conf
        if conf:
            self.verticalHeader().setSectionsMovable(True)
            self.verticalHeader().show()
        else:
            self.verticalHeader().setSectionsMovable(False)
            self.verticalHeader().hide()

    @QtCore.Slot(bool)
    def request_configurable(self, conf: bool):
        """
        Designable slot for toggling config mode.
        """
        self.configurable = conf


class ChannelTableWidgetItem(QtWidgets.QTableWidgetItem):
    """
    QTableWidgetItem that gets values from a PyDMChannel

    Parameters
    ----------
    header : str
        The name of the header of the column
    default : any, optional
        Starting value for the cell
    channel : str, optional
        PyDM channel address for value and connection updates.
    deadband : float, optional
        Only update the table if the change is more than the deadband.
        This can help make large tables less resource-hungry.
    """
    header: str
    channel: str | None
    deadband: float
    pydm_channel: PyDMChannel | None

    def __init__(
        self,
        header: str,
        default: Any | None = None,
        channel: str | None = None,
        deadband: float = 0.0,
        parent: QtWidgets.QWidget | None = None
    ):
        super().__init__(parent)
        self.header = header
        self.update_value(default)
        self.channel = channel
        self.deadband = deadband
        if channel is None:
            self.update_connection(True)
            self.pydm_channel = None
        else:
            self.update_connection(False)
            self.pydm_channel = PyDMChannel(
                channel,
                value_slot=self.update_value,
                connection_slot=self.update_connection,
            )
            self.pydm_channel.connect()

    def update_value(self, value: Any) -> str:
        """
        Store the value for sorting and display in the table if visible.

        By setting the text, we also notify the table that a cell has updated.
        """
        try:
            if abs(self._value - value) < self.deadband:
                return
        except Exception:
            pass
        self._value = value
        self.setText(str(value))

    def update_connection(self, connected: bool) -> None:
        """
        When our PV connects or disconnects, store the state as an attribute.
        """
        self.connected = connected

    def get_value(self) -> Any:
        return self._value

    def __lt__(self, other: ChannelTableWidgetItem) -> bool:
        """
        Two special sorting rules:
        1. None is the greatest
        2. Empty string is the greatest string

        This means that disconnected and empty string sort as "high"
        (sort ascending is most common)
        """
        # Make sure None sorts as greatest
        if self.get_value() is None:
            return False
        elif other.get_value() is None:
            return True
        # Make sure empty string sorts as next greatest
        elif self.get_value() == '':
            return False
        elif other.get_value() == '':
            return True
        return self.get_value() < other.get_value()


@dataclasses.dataclass
class FilterInfo:
    filter_func: Callable[[dict[str, Any]], bool]
    active: bool
    name: str



from qtpy.QtCore import QPointF
from qtpy.QtGui import QPainterPath

from .base import BaseSymbolIcon


class RGASymbolIcon(BaseSymbolIcon):
    """
    A widget with a residual gas analyzer symbol drawn in it.

    Parameters
    ----------
    parent : QWidget
        The parent widget for the icon
    """
    path = QPainterPath(QPointF(0, 0))
    path.lineTo(1, 0)
    path.lineTo(1, 0.33)
    path.lineTo(0.33, 0.33)
    path.lineTo(0.33, 0.66)
    path.lineTo(0.33, 0.66)
    path.lineTo(1, 0.66)
    path.lineTo(1, 1)
    path.lineTo(0, 1)
    path.closeSubpath()

    def draw_icon(self, painter):
        painter.drawPath(self.path)
